# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
cimport cython

import numpy as np
cimport numpy as np

from marsi.chemistry.common_ext cimport _tanimoto_distance

ctypedef np.int32_t INT32_t
ctypedef np.float32_t FLOAT32_t


cdef extern int __builtin_popcount(unsigned int) nogil

@cython.nonecheck(False)
@cython.cdivision(True)
cdef class CNearestNeighbors:
    cdef np.ndarray _features
    cdef np.ndarray _features_lengths
    cdef np.ndarray _start_positions

    def __init__(self, features, features_lengths):
        self._features = features
        self._features_lengths = features_lengths
        self._start_positions = None

    def __getstate__(self):
        return {
            '_features': self._features,
            '_features_lengths': self._features_lengths,
            '_start_positions': self._start_positions
        }

    def __setstate__(self, state):
        self._features = state['_features']
        self._features_lengths = state['_features_lengths']
        self._start_positions = state['_start_positions']

    @property
    def features(self):
        return self._features

    @property
    def features_lengths(self):
        return self._features_lengths

    @property
    def start_positions(self):
        if self._start_positions is None:
            self._start_positions = self._calc_start_positions()
        return self._start_positions

    def distances_py(self, fingerprint):
        return self._distances(fingerprint)

    cdef np.ndarray[INT32_t, ndim=1] _calc_start_positions(self):
        cdef unsigned int i = 0
        cdef unsigned int n = len(self._features_lengths)

        cdef np.ndarray[INT32_t, ndim=1] res = np.empty(n, dtype=np.int32)

        res[0] = i
        for pos in range(n-1):
            i += self._features_lengths[pos]
            res[pos+1] = i

        return res

    cdef np.ndarray[FLOAT32_t, ndim=1] _distances(self, np.ndarray[INT32_t, ndim=1] fingerprint):
        cdef unsigned int n = len(self._features_lengths)
        cdef np.ndarray[FLOAT32_t, ndim=1] _distances = np.zeros(n, dtype=np.float32)

        for i in range(n):
            start = self.start_positions[i]
            size = self._features_lengths[i]
            _distances[i] = _tanimoto_distance(fingerprint, self._features[start:start+size])

        return _distances




# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cameo import pfba, load_model
from cameo.strain_design.heuristic.evolutionary.objective_functions import biomass_product_coupled_yield
from cement.core.controller import CementBaseController, expose
from pandas import DataFrame

from marsi.utils import BIOMASS_RE
from marsi.cobra.strain_design import RandomMutagenesisDesign, ALEDesign
from marsi.cobra.utils import CURRENCY_METABOLITES
from marsi.cobra.strain_design.post_processing import replace_design
from marsi.utils import default_carbon_sources, unique


class OptimizationController(CementBaseController):
    """
    This is the Optimization Controller. It allows to run optimizations from the command line.

    """
    exclude_reactions = ["ATPM"]

    class Meta:
        label = 'optimize'
        stacked_on = 'base'
        stacked_type = 'nested'
        description = "Optimize a host phenotype using Metabolites as targets"
        arguments = [
            (['--model', '-m'], dict(help="path or identifier of the model to be used")),
            (['--carbon-source'], dict(help="(optional) the carbon source exchange reaction. "
                                            "It will be auto-detected if not defined")),
            (['--target', '-t'], dict(help="The exchange reaction of the target phenotype or "
                                           "the metabolite to accumulate")),
            (['--approach', '-a'], dict(help="classic: use reaction search; metabolites: search for "
                                             "metabolite targets directly. (default: classic)")),
            (['--biomass'], dict(help="(optional) the biomass reaction (needed if there is none or many reactions "
                                      "containing biomass on their ids or names")),
            (['--max-interventions', '-mi'], dict(help="(optional) maximum number of interventions (default: 2)")),
            (['--max-evaluations', '-me'], dict(help='(optional) maximum number of evaluations (default: 20000)')),
            (['--output-file', '-o'], dict(help="output file")),
            (['--input-file', '-i'], dict(help="input file")),
            (['--transporters'], dict(help="Include transporters knockout"))
        ]

    @expose(hide=True)
    def default(self):
        pass

    @expose("Run optimization assuming Chemical Mutagenesis scenario")
    def mutagenesis(self):
        """
        1. Load model specified by --model.
        2. Detect carbon source if not defined, otherwise change the carbon source.
        3. Detect biomass if not defined
        4. Run the optimization.
        5. Return the results.
        """

        model, target, carbon_source, biomass, output_file = self._parse_optimization_args()

        designer = RandomMutagenesisDesign(model=model)

        max_interventions = 2
        max_evaluations = 20000

        if self.app.pargs.max_interventions is not None:
            max_interventions = int(self.app.pargs.max_interventions)

        if self.app.pargs.max_evaluations is not None:
            max_evaluations = int(self.app.pargs.max_evaluations)

        if self.app.pargs.transporters is not None:
            transporters = [r.id for r in model.reactions if len(set(m.compartment for m in r.metabolites)) > 1]
            exclude_reactions = self.exclude_reactions
            exclude_reactions += transporters
        else:
            exclude_reactions = self.exclude_reactions

        print("---------------------------------------------------")
        print("Chemical Mutagenesis Optimization setup:")
        print("- Search method: %s" % self.app.pargs.approach)
        print("- Target: %s" % target.id)
        print("- Biomass: %s" % biomass.id)
        print("- Carbon source: %s" % carbon_source.id)
        print("- Max evaluations: %i" % max_evaluations)
        print("- Max interventions: %i" % max_interventions)

        if self.app.pargs.approach == 'classic':
            result = designer.optimize_with_reaction(target=target, substrate=carbon_source, biomass=biomass,
                                                     design_method="optgene", manipulation_type="reactions",
                                                     max_interventions=max_interventions,
                                                     essential_reactions=exclude_reactions,
                                                     max_evaluations=max_evaluations)
        else:
            result = designer.optimize_with_metabolites(target=target, substrate=carbon_source, biomass=biomass,
                                                        max_interventions=max_interventions,
                                                        max_evaluations=max_evaluations)

        result.to_csv(output_file)

        print("Finished")

    @expose(help="Run optimization assuming Adaptive Laboratory Evolution scenario")
    def ale(self):
        """
                1. Load model specified by --model.
                2. Detect carbon source if not defined, otherwise change the carbon source.
                3. Detect biomass if not defined
                4. Run the optimization.
                5. Return the results.
                """

        model, target, carbon_source, biomass, output_file = self._parse_optimization_args()

        designer = ALEDesign(model=model)

        max_interventions = 2
        max_evaluations = 20000

        if self.app.pargs.max_interventions is not None:
            max_interventions = int(self.app.pargs.max_interventions)

        if self.app.pargs.max_evaluations is not None:
            max_evaluations = int(self.app.pargs.max_evaluations)

        if self.app.pargs.transporters is not None:
            transporters = [r.id for r in model.reactions if len(set(m.compartment for m in r.metabolites)) > 1]
            exclude_reactions = self.exclude_reactions
            exclude_reactions += transporters
        else:
            exclude_reactions = self.exclude_reactions

        print("---------------------------------------------------")
        print("ALE Optimization setup:")
        print("- Search method: %s" % self.app.pargs.approach)
        print("- Target: %s" % target.id)
        print("- Biomass: %s" % biomass.id)
        print("- Carbon source: %s" % carbon_source.id)
        print("- Max evaluations: %i" % max_evaluations)
        print("- Max interventions: %i" % max_interventions)

        if self.app.pargs.approach == 'classic':
            result = designer.optimize_with_reaction(target=target, substrate=carbon_source, biomass=biomass,
                                                     design_method="differential_fva", manipulation_type="reactions",
                                                     max_interventions=max_interventions,
                                                     essential_reactions=exclude_reactions)
        else:
            result = designer.optimize_with_metabolites(target=target, substrate=carbon_source, biomass=biomass,
                                                        max_interventions=max_interventions,
                                                        max_evaluations=max_evaluations)

        result.to_csv(output_file)

        print("Finished")

    @expose(help="Convert output from OptGene or OptKnock to Anti-metabolite approach")
    def convert(self):
        """
        1. Load model specified by --model.
        2. Detect carbon source if not defined, otherwise change the carbon source.
        3. Detect biomass if not defined
        4. Convert the results.

        Returns
        -------

        """
        input_file = None

        if self.app.pargs.input_file is None:
            print("--input-file is required")
            exit(1)
        else:
            input_file = self.app.pargs.input_file

        knockouts = DataFrame.from_csv(input_file)
        model, target, carbon_source, biomass, output_file = self._parse_optimization_args()

        objective_function = biomass_product_coupled_yield(biomass, target, carbon_source)

        knockouts['designs'] = knockouts.reactions.apply(lambda s: s)

        result = replace_design(model, knockouts, objective_function, pfba, {}, CURRENCY_METABOLITES)
        result.to_csv(output_file)

        print("Finished")

    def _parse_optimization_args(self):
        target = None
        model = None
        carbon_source = None
        output_file = None

        if self.app.pargs.output_file is None:
            print("--output-file is required")
            exit(1)
        else:
            output_file = self.app.pargs.output_file

        if self.app.pargs.model is None:
            print("--model is required")
            exit(1)
        else:
            try:
                model = load_model(self.app.pargs.model)
            except Exception:
                print("Invalid model %s" % self.app.pargs.model)
                exit(1)

        if self.app.pargs.target is None:
            print("--target is required")
            exit(1)
        else:
            try:
                target = model._reaction_for(self.app.pargs.target)
            except KeyError:
                print("Invalid target %s" % self.app.pargs.target)
                exit(1)

        if self.app.pargs.approach is None:
            self.app.pargs.approach = 'classic'

        model_carbon_source = default_carbon_sources(model)[0]
        if self.app.pargs.carbon_source is not None:
            try:
                carbon_source = model._reaction_for(self.app.pargs.carbon_source)
            except KeyError:
                print("Invalid carbon source %s" % self.app.pargs.carbon_source)
                exit(1)

            carbon_source.lower_bound = model_carbon_source.lower_bound
            model_carbon_source.lower_bound = 0
        else:
            carbon_source = model_carbon_source

        if self.app.pargs.biomass is not None:
            biomass = model.reactions.get_by_id(self.app.pargs.biomass)
        else:
            def is_biomass(name):
                if len(BIOMASS_RE.findall(name)) > 0:
                    return True
                else:
                    return False

            biomass = list(model.reactions.query(is_biomass, "id"))
            biomass += list(model.reactions.query(is_biomass, "name"))
            unique(biomass)

            if len(biomass) == 0:
                print("Cannot find biomass reaction")
                exit(1)
            elif len(biomass) > 1:
                print("Multiple biomass reactions found!")
                print("\n".join("%i: %s" % (i + 1, b.id) for i, b in enumerate(biomass)))
                choice = input("Please chose: ")
                try:
                    choice = int(choice)
                    biomass = biomass[choice - 1]
                except (IndexError, TypeError):
                    print("Invalid choice %i" % choice)
                    exit(1)
            else:
                biomass = biomass[0]

        return model, target, carbon_source, biomass, output_file



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import pubchempy
from bioservices.chebi import ChEBI
from bioservices.kegg import KEGG
from bioservices.uniprot import UniProt

from cachetools import cached, LRUCache
from cachetools.keys import hashkey

from marsi.chemistry.openbabel import mol_str_to_inchi

lru_cache = LRUCache(maxsize=1024)


try:
    __all__ = ['find_chebi_id', 'find_pubchem_id', "map_uniprot_from_pdb_ids", "find_best_chebi_structure"]

    chebi_client = ChEBI()
    kegg_client = KEGG()
    uniprot_client = UniProt()

    @cached(lru_cache, key=partial(hashkey, 'uniprot'))
    def map_uniprot_from_pdb_ids(pdb_ids):
        return uniprot_client.mapping(fr="PDB_ID", to="ACC", query=pdb_ids)

    @cached(lru_cache, key=partial(hashkey, 'uniprot'))
    def inchi_from_chebi(chebi_id):
        try:
            return chebi_client.getCompleteEntity(chebi_id).inchi.strip()
        except AttributeError:
            return None

    @cached(lru_cache, key=partial(hashkey, 'uniprot'))
    def inchi_from_kegg(kegg_id):
        try:
            return mol_str_to_inchi(kegg_client.get(kegg_id, 'mol'))
        except:
            return None

    def find_chebi_id(metabolite):
        """
        Queries ChEBI using InChI Key.

        Parameters
        ----------
        metabolite : Metabolite

        Returns
        -------
        list
            With tuples of (id, name)
        """
        try:
            query = chebi_client.getLiteEntity(metabolite.inchi_key, searchCategory="INCHI KEY")
            return [(e.chebiId, e.chebiAsciiName) for e in query][0]
        except IndexError:
            raise KeyError("%s not found in ChEBI" % metabolite.inchi_key)

    def find_pubchem_id(metabolite):
        """
        Queries PubChem Compound using InChI Key.

        Parameters
        ----------
        metabolite : Metabolite

        Returns
        -------
        list
            With tuples of (id, name)
        """
        try:
            return [(str(e.cid), e.iupac_name) for e in pubchempy.get_compounds(metabolite.inchi_key,
                                                                                namespace="inchikey")][0]
        except:
            raise KeyError("%s not found in PubChem Compound" % metabolite.inchi_key)

except Exception:
    from warnings import warn
    __all__ = ['no_services_available', 'find_best_chebi_structure']
    no_services_available = "Please check your internet connection"

    def map_uniprot_from_pdb_ids(pdb_ids):
        raise RuntimeError(no_services_available)

    def inchi_from_chebi(chebi_id):
        raise RuntimeError(no_services_available)

    def inchi_from_kegg(kegg_id):
        raise RuntimeError(no_services_available)

    def find_chebi_id(metabolite):
        raise RuntimeError(no_services_available)

    def find_pubchem_id(metabolite):
        raise RuntimeError(no_services_available)

    warn(no_services_available)


def find_best_chebi_structure(entity):
    structure = entity.ChemicalStructures[0]
    for i, struct in enumerate(entity.ChemicalStructures):
        if struct.dimension == "3D":
            if structure.dimension == "3D" and structure.defaultStructure:
                return structure.structure
            else:
                structure = struct
        else:
            if struct.defaultStructure and not structure.dimension == "3D":
                structure = struct

    return structure.structure



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.optimize import curve_fit
from marsi.processing.chemistry.openbabel import smiles_to_molecule


def rss(measured, predicted):
    assert isinstance(measured, np.ndarray)
    assert isinstance(predicted, np.ndarray)

    residuals = (measured - predicted)**2
    return sum(residuals)


def fit_model(measured, smiles):
    molecules = [smiles_to_molecule(_smiles) for _smiles in smiles]
    return curve_fit(calc_solubility, molecules, measured)


def calc_solubility(molecules, i, log_p, mw, hbd, hba, mp, mr, abonds, rb, ap):
    y = [0 for _ in molecules]

    for j, molecule in enumerate(molecules):
        desc = molecule.calcdesc()

        rotatable_bounds = molecule.OBMol.NumRotors()
        aromatic_partition = len([a for a in molecule.atoms if a.OBAtom.IsAromatic()]) / desc['atoms']

        y[j] = i + log_p * desc['logP'] + mw * desc['MW'] + hbd * desc['HBD'] + hba * desc['HBA2'] + \
               mp * desc['MP'] + mr * desc['MR'] + abonds * desc['abonds'] + rb * rotatable_bounds + \
               ap * aromatic_partition

    return np.array(y)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Quantitative Structure-Activity Relationship.

This module contains basic Machine Learning stuff to analyse the IC50 data:
    1. Feature extraction;
    2. Model training;
    3. Feature analysis;
"""
import numpy
from pandas import Series, DataFrame

from marsi.processing.chemistry.openbabel import get_spectrophore_data


def _build_feature_row(metabolite, *functions, **params):
    return Series(name=metabolite.inchi_key,
                  data=[f(metabolite, **params) for f in functions],
                  index=[f.__name__ for f in functions])


def build_custom_feature_table(metabolites, *functions, **params):
    result = DataFrame()

    for metabolite in metabolites:
        result = result.append(_build_feature_row(metabolite, *functions, **params))

    return result


def build_target_table(ic50s):
    result = DataFrame(columns=["inchi_key", "target_name", "target_organism", "ic50"])
    for i, ic50 in enumerate(ic50s):
        result.loc[i] = [ic50.metabolite.inchi_key, ic50.target_name, ic50.target_organism, ic50.value]

    return result


def build_spectrophore_feature_table(metabolites):
    result = DataFrame(columns=["spectrophore_feature_%i" % (i + 1) for i in range(48)])
    for metabolite in metabolites:
        mol = metabolite.molecule(library='openbabel')
        mol.make3D()
        result.loc[metabolite.inchi_key] = get_spectrophore_data(mol)

    return result


def build_maccs_feature_table(metabolites):
    result = DataFrame(columns=["maccs%i" % (i + 1) for i in range(165)])
    for metabolite in metabolites:
        hashed = metabolite.fingerprint('maccs', hash=True, bits=165)
        result.loc[metabolite.inchi_key] = numpy.array(hashed, dtype=int)

    return result


def build_table(ic50s, use_maccs=True, use_spectrophore=True, *functions, **kwargs):
    target_table = build_target_table(ic50s)
    metabolites = set(ic50.metabolite for ic50 in ic50s)
    if use_maccs:
        print("Building maccs feature table")
        maccs_table = build_maccs_feature_table(metabolites)
        target_table = target_table.merge(maccs_table, right_index=True, left_on='inchi_key')
    if use_spectrophore:
        print("Building spectrophore feature table")
        spectrophore_table = build_spectrophore_feature_table(metabolites)
        target_table = target_table.merge(spectrophore_table, right_index=True, left_on='inchi_key')
    if len(functions) > 0:
        print("Building custom feature table (%in functions)" % len(functions))
        custom_table = build_custom_feature_table(metabolites, *functions, **kwargs)
        target_table = target_table.merge(custom_table, right_index=True, left_on='inchi_key')

    return target_table



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cement.core.controller import CementBaseController, expose


from marsi.chemistry.molecule import Molecule
from marsi.io import write_excel_file
from marsi.nearest_neighbors import search_closest_compounds

OUTPUT_WRITERS = {
    'csv': lambda df, path, *args: df.to_csv(path),
    'excel': lambda df, path, *args: write_excel_file(df, path)

}


class ChemistryController(CementBaseController):
    """
    This is the Optimization Controller. It allows to run optimizations from the command line.

    """

    class Meta:
        label = 'chem'
        stacked_on = 'base'
        stacked_type = 'nested'
        description = "Metabolite database query and filter tools"
        arguments = [
            (['--inchi'], dict(help="The metabolite InChI to search")),
            (['--sdf'], dict(help="The metabolite SDF to search")),
            (['--mol'], dict(help="The metabolite MOL to search")),
            (['--fingerprint-format', '-fp'], dict(help="The fingerprint format", default='maccs', action='store')),
            (['--neighbors', '-k'], dict(help="Filter the first K hits")),
            (['--radius', '-r'], dict(help="Filter hits within R distance radius")),
            (['--atoms-weight', '-aw'], dict(help="The weight of the atoms for structural similarity")),
            (['--bonds-weight', '-bw'], dict(help="The weight of the bonds for structural similarity")),
            (['--atoms-diff', '-ad'], dict(help="The maximum number of the atoms difference for database query")),
            (['--bonds-diff', '-bd'], dict(help="The maximum number of the bonds difference for database query")),
            (['--rings-diff', '-rd'], dict(help="The maximum number of the rings difference for database query")),
            (['--output-file', '-o'], dict(help="Output file")),
            (['--output-format', '-f'], dict(help='Output format (default: csv)', default='csv'))
        ]

    @expose(hide=True)
    def default(self):
        print("Welcome to MARSI chemistry package")
        print("Here you can find the tools to find and sort analogs for metabolites")

    @expose(help="Find analogs for a metabolite")
    def find_analogs(self):
        """
        1. Make a fingerprint from the --inchi, --sdf or --mol.
        2. Query the database

        Returns
        -------

        """
        output_file = None

        if self.app.pargs.output_file is not None:
            output_file = self.app.pargs.output_file
        else:
            print("--output-file argument is required")
            exit(1)

        output_file += ".%s" % self.app.pargs.output_format

        molecule = None

        bonds_weight = 0.5
        atoms_weight = 0.5
        bonds_diff = 3
        atoms_diff = 4
        rings_diff = 3

        if self.app.pargs.bonds_weight is not None:
            bonds_weight = float(self.app.pargs.bonds_weight)

        if self.app.pargs.atoms_weight is not None:
            atoms_weight = float(self.app.pargs.atoms_weight)

        if self.app.pargs.bonds_diff is not None:
            bonds_diff = float(self.app.pargs.bonds_diff)

        if self.app.pargs.atoms_diff is not None:
            atoms_diff = float(self.app.pargs.atoms_diff)

        if self.app.pargs.rings_diff is not None:
            rings_diff = float(self.app.pargs.rings_diff)

        if self.app.pargs.inchi is not None:
            molecule = Molecule.from_inchi(self.app.pargs.inchi)
        elif self.app.pargs.sdf is not None:
            molecule = Molecule.from_sdf(self.app.pargs.sdf)
        elif self.app.pargs.mol is not None:
            molecule = Molecule.from_mol(self.app.pargs.mol)
        else:
            print("Please provide one of the following inputs --inchi, --sdf or --mol")
            exit(1)

        fpformat = self.app.pargs.fingerprint_format

        results = search_closest_compounds(molecule=molecule, fpformat=fpformat, bonds_weight=bonds_weight,
                                           bonds_diff=bonds_diff, atoms_weight=atoms_weight, atoms_diff=atoms_diff,
                                           rings_diff=rings_diff)
        results.sort_values('structural_similarity', ascending=False, inplace=True)
        OUTPUT_WRITERS[self.app.pargs.output_format](results, output_file, None)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from pandas import DataFrame


def parse_kegg_brite(brite_file):
    kegg = DataFrame(columns=['group', 'family', 'level', 'target', 'generic_name',
                              'name', 'drug_type', 'kegg_drug_id'])

    with open(brite_file) as kegg_data:
        group = None
        family = None
        generic_name = None
        level = None
        target = None
        i = 0
        for line in kegg_data:
            line = line.strip("\n")
            if line.startswith("A"):
                group = line[1:].strip("<b>").strip("<b/>")
            if group != "Enzymes":
                continue
            else:
                if line.startswith("B"):
                    family = line[1:].strip()
                    level = family
                elif line.startswith("C"):
                    target = line[1:].strip()
                elif line.startswith("D"):
                    generic_name = line[1:].strip()
                elif line.startswith("E"):
                    line = line[1:].strip()
                    split = line.split()
                    name = " ".join(split[1:-2])
                    kegg.loc[i] = [group, family, level, target, generic_name, name, split[-1], split[0]]
                    i += 1

    print("Found %i drugs acting on enzymes" % i)
    return kegg


def parse_chebi_data(chebi_names_file, chebi_vertice_file, chebi_relation_file):
    chebi_names = DataFrame.from_csv(chebi_names_file, sep="\t")
    chebi_names.fillna("", inplace=True)
    chebi_names.index.name = "id"

    chebi_names.columns = map(str.lower, chebi_names.columns)
    chebi_names.drop_duplicates('compound_id', keep='last', inplace=True)
    chebi_names['adapted'] = chebi_names.adapted.apply(lambda v: v == "T")

    chebi_analogues = chebi_names[chebi_names.name.str.contains('analog')]
    chebi_antimetabolite = chebi_names[chebi_names.compound_id == 35221]

    chebi_relations = DataFrame.from_csv(chebi_relation_file, sep="\t")
    chebi_relations.columns = map(str.lower, chebi_relations.columns)
    chebi_relations.index.name = "id"

    chebi_vertices = DataFrame.from_csv(chebi_vertice_file, sep="\t")
    chebi_vertices.columns = map(str.lower, chebi_vertices.columns)
    chebi_vertices.index.name = "id"

    def retrieve_child_id(compound_id):
        return chebi_vertices.loc[compound_id, 'compound_child_id']

    chebi_relations['init_compound_id'] = chebi_relations.init_id.apply(retrieve_child_id)
    chebi_relations['final_compound_id'] = chebi_relations.final_id.apply(retrieve_child_id)

    chebi_is_a = chebi_relations[chebi_relations['type'] == 'is_a']
    chebi_has_role = chebi_relations[chebi_relations['type'] == 'has_role']

    def recursive_search(roots, relations, universe, aggregated, forward=True):
        aggregated = aggregated.append(roots, ignore_index=True)
        if forward:
            filtered = relations[relations.init_compound_id.isin(roots.compound_id)]
            roots = universe[universe.compound_id.isin(filtered.final_compound_id)]
        else:
            filtered = relations[relations.final_compound_id.isin(roots.compound_id)]
            roots = universe[universe.compound_id.isin(filtered.init_compound_id)]

        if len(roots) > 0:
            aggregated, roots = recursive_search(roots, relations, universe, aggregated, forward)

        return aggregated, roots

    data = DataFrame(columns=chebi_names.columns)
    anti = DataFrame(columns=chebi_names.columns)

    data, _ = recursive_search(chebi_analogues, chebi_is_a, chebi_names, data, True)
    data, _ = recursive_search(chebi_antimetabolite, chebi_is_a, chebi_names, data, True)

    anti, _ = recursive_search(chebi_antimetabolite, chebi_has_role, chebi_names, anti, True)
    data, _ = recursive_search(anti, chebi_is_a, chebi_names, data, True)

    data['compound_id'] = data.compound_id.apply(int)
    return data


def parse_pubchem(summary_file):
    pubchem = DataFrame(columns=["name", "molecular_weight", "formula", "uipac_name", "create_date", "compound_id"])

    with open(summary_file) as pubchem_data:
        row = dict(name=None, molecular_weight=None, formula=None, uipac_name=None,
                   create_date=None, compound_id=None)
        i = 0
        for line in pubchem_data:
            line = line.strip("\n")
            if len(line) == 0:
                if any(v for v in row.values()):
                    pubchem.loc[i] = [row[k] for k in pubchem.columns]
                    i += 1
                row = dict(name=None, molecular_weight=None, formula=None,
                           uipac_name=None, create_date=None, compound_id=None)
            elif re.match("^\d+\.*", line):
                row['name'] = line.split(". ", 1)[1].split("; ")[0]
            elif line.startswith("MW:"):
                match = re.match("MW:\s+(\d+\.\d+).+MF:\s(\w+)", line)
                row['molecular_weight'] = float(match.group(1))
                row['formula'] = match.group(2)
            elif line.startswith("IUPAC name:"):
                row['uipac_name'] = line[10:]
            elif line.startswith("Create Date:"):
                row['create_date'] = line[12:]
            elif line.startswith("CID:"):
                row['compound_id'] = int(line[5:])

    pubchem['compound_id'] = pubchem.compound_id.apply(int)
    return pubchem



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from cameo.flux_analysis.simulation import FluxDistributionResult
from cobra.core.dictlist import DictList
from cobra.core.model import Model
from cobra.core.reaction import Reaction
from pandas import Series


__all__ = ["compete_metabolite", "inhibit_metabolite", "knockout_metabolite", "apply_anti_metabolite"]

logger = logging.getLogger(__name__)


def compete_metabolite(model, metabolite, reference_dist, fraction=0.5, allow_accumulation=True, constant=1e4):
    """
    Increases the usage of a metabolite based on a reference flux distributions.

    Parameters
    ----------
    model : Model
        A constraint-based model.
    metabolite : cobra.Metabolite
        A metabolite.
    reference_dist : dict or FluxDistributionResult
        The result of a FBA like simulation. Alternative can be dictionaries of reaction.id -> flux.
    fraction : float
        How much does it requires the reactions to go up.
    allow_accumulation : bool
        Allow to accumulate the metabolite (add a exchange reaction).
    constant : float
        A large number (like 10000).

    Returns
    -------
    cobra.core.Reaction
        If allow accumulation returns the exchange reaction associated with the metabolite.
    """
    assert isinstance(model, Model)

    reactions = [r for r in metabolite.reactions if len(set(m.compartment for m in r.metabolites)) == 1]

    if isinstance(reference_dist, FluxDistributionResult):
        reference_dist = reference_dist.fluxes.to_dict()
    elif isinstance(reference_dist, Series):
        reference_dist = reference_dist.to_dict()

    if not isinstance(reference_dist, dict):
        raise ValueError("'reference_dist' must be a dict or FluxDistributionResult")

    exchanges = DictList(model.exchanges)

    exchange = None
    if allow_accumulation:
        species_id = metabolite.id[:-2]
        if "EX_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("EX_%s_e" % species_id)
        elif "DM_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("DM_%s_e" % species_id)
        else:
            reaction_id = "COMPETE_%s" % metabolite.id
            if reaction_id in model.reactions:
                exchange = model.reactions.get_by_id(reaction_id)
            else:
                exchange = model.add_boundary(metabolite, type="compete sink", reaction_id=reaction_id, lb=0)

    aux_variables = {}
    ind_variables = {}
    turnover = sum(abs(r.metabolites[metabolite] * reference_dist.get(r.id, 0)) for r in metabolite.reactions)
    for reaction in reactions:
        coefficient = reaction.metabolites[metabolite]

        # Optimization to reduce y variables and problem complexity:
        # Irreversible reactions that only produce the metabolite can be ignored because they will not contribute
        # to the consumption turnover. Reactions that only consume the metabolite can be added directly into the
        # sum constraint. This allows for a less complex problem with less variables.

        if not reaction.reversibility:
            if coefficient < 0:  # skip reactions that can only produce the metabolite
                continue
            else:  # keep the v*coefficient value for reactions that can only consume the metabolite
                aux_variables[reaction.id] = reaction.flux_expression * coefficient
                continue

        to_add = []

        ind_var_id = "y_%s" % reaction.id
        aux_var_id = "u_%s" % reaction.id
        try:
            ind_var = model.solver.variables[ind_var_id]
            aux_var = model.solver.variables[aux_var_id]
        except KeyError:
            ind_var = model.solver.interface.Variable(ind_var_id, type='binary')
            aux_var = model.solver.interface.Variable(aux_var_id, lb=0)
            to_add += [ind_var, aux_var]

        aux_variables[reaction.id] = aux_var
        ind_variables[reaction.id] = ind_var

        upper_indicator_constraint_name = "ind_%s_u" % reaction.id
        lower_indicator_constraint_name = "ind_%s_l" % reaction.id

        auxiliary_constraint_a_name = "aux_%s_a" % reaction.id
        auxiliary_constraint_b_name = "aux_%s_b" % reaction.id
        auxiliary_constraint_c_name = "aux_%s_c" % reaction.id
        auxiliary_constraint_d_name = "aux_%s_d" % reaction.id

        try:
            model.solver.constraints[upper_indicator_constraint_name]
        except KeyError:

            # constraint y to be 0 if Sv >= 0 (production)

            #   -M             0                M
            # Sv <-------------|---------------->
            #         y=0      |      y=1

            # Sv - My <= 0
            # if y = 1 then Sv <= M
            # if y = 0 then Sv <= 0
            upper_indicator_expression = coefficient * reaction.flux_expression - ind_var * constant
            ind_constraint_u = model.solver.interface.Constraint(upper_indicator_expression,
                                                                 name=upper_indicator_constraint_name,
                                                                 ub=0)

            # Sv + M(1-y) >= 0
            # if y = 1 then Sv >= 0
            # if y = 0 then Sv >= -M
            lower_indicator_expression = coefficient * reaction.flux_expression + constant - ind_var * constant
            ind_constraint_l = model.solver.interface.Constraint(lower_indicator_expression,
                                                                 name=lower_indicator_constraint_name,
                                                                 lb=0)

            # a) -My + u <= 0
            # b) My + u >= 0

            # if y = 0, u = 0
            # if y = 1, -M <= u <= M
            aux_indicator_expression_a = -constant * ind_var + aux_var
            aux_constraint_a = model.solver.interface.Constraint(aux_indicator_expression_a,
                                                                 name=auxiliary_constraint_a_name,
                                                                 ub=0)

            aux_indicator_expression_b = constant * ind_var + aux_var
            aux_constraint_b = model.solver.interface.Constraint(aux_indicator_expression_b,
                                                                 name=auxiliary_constraint_b_name,
                                                                 lb=0)
            #
            # # c) -M(1-y) + u - viSi <= 0
            # # d) M(1-y) + u - viSi >= 0
            #
            # # if y = 1 then 0 <= u - viSi <= 0
            # # if y = 0 then -M <= u - viSi <= M
            aux_indicator_expression_c = -constant * (1 - ind_var) + aux_var - reaction.flux_expression * coefficient
            aux_constraint_c = model.solver.interface.Constraint(aux_indicator_expression_c,
                                                                 name=auxiliary_constraint_c_name,
                                                                 ub=0)

            aux_indicator_expression_d = constant * (1 - ind_var) + aux_var - reaction.flux_expression * coefficient
            aux_constraint_d = model.solver.interface.Constraint(aux_indicator_expression_d,
                                                                 name=auxiliary_constraint_d_name,
                                                                 lb=0)

            to_add += [ind_constraint_l, ind_constraint_u, aux_constraint_a,
                       aux_constraint_b, aux_constraint_c, aux_constraint_d]

            model.add_cons_vars(to_add)

    min_production_turnover = (1 + fraction) * (turnover / 2)
    # sum(u) >= (1 + fraction) * uWT
    constrain_name = "take_less_%s" % metabolite.id
    if constrain_name not in model.constraints:
        increase_turnover_constraint = model.solver.interface.Constraint(sum(aux_variables.values()),
                                                                         name="take_less_%s" % metabolite.id,
                                                                         lb=min_production_turnover)

        model.add_cons_vars(increase_turnover_constraint)
    else:
        increase_turnover_constraint = model.constraints[constrain_name]
        increase_turnover_constraint.lb = min_production_turnover

    return exchange


def inhibit_metabolite(model, metabolite, reference_dist, fraction=0.5, allow_accumulation=True, constant=1e4):
    """
    Inhibits the usage of a metabolite based on a reference flux distributions.

    Parameters
    ----------
    model : Model
        A constraint-based model.
    metabolite : cobra.Metabolite
        A metabolite.
    reference_dist : dict, FluxDistributionResult
        The result of a FBA like simulation. Alternative can be dictionaries of reaction.id -> flux.
    fraction : float
        How much does it inhibits the reactions. A float applies the same amount of inhibition. A dictionary must
        contain an inhibition percentage to all reactions associated with the metabolite.
    allow_accumulation : bool
        Allow to accumulate the metabolite (add a exchange reaction).
    constant : float
        A large number (like 10000).

    Returns
    -------
    cameo.core.Reaction, None
        If allow accumulation returns the exchange reaction associated with the metabolite.

    """
    reactions = [r for r in metabolite.reactions if len(set(m.compartment for m in r.metabolites)) == 1]

    if isinstance(reference_dist, FluxDistributionResult):
        reference_dist = reference_dist.fluxes.to_dict()
    elif isinstance(reference_dist, Series):
        reference_dist = reference_dist.to_dict()

    if not isinstance(reference_dist, dict):
        raise ValueError("'reference_dist' must be a dict or FluxDistributionResult")

    exchanges = DictList(model.exchanges)

    exchange = None

    if allow_accumulation:
        species_id = metabolite.id[:-2]
        if "EX_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("EX_%s_e" % species_id)
        elif "DM_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("DM_%s_e" % species_id)
        else:
            reaction_id = "INHIBIT_%s" % metabolite.id
            if reaction_id in model.reactions:
                exchange = model.reactions.get_by_id(reaction_id)
            else:
                exchange = model.add_boundary(metabolite, type="inhibit sink", reaction_id=reaction_id, lb=0)

    aux_variables = {}
    ind_variables = {}
    turnover = sum(abs(r.metabolites[metabolite] * reference_dist.get(r.id, 0)) for r in metabolite.reactions)
    for reaction in reactions:
        coefficient = reaction.metabolites[metabolite]

        # Optimization to reduce y variables and problem complexity:
        # Irreversible reactions that only produce the metabolite can be ignored because they will not contribute
        # to the consumption turnover. Reactions that only consume the metabolite can be added directly into the
        # sum constraint. This allows for a less complex problem with less variables.

        if not reaction.reversibility:
            if coefficient > 0:  # skip reactions that can only produce the metabolite
                continue
            else:  # keep the v*coefficient value for reactions that can only consume the metabolite
                aux_variables[reaction.id] = - reaction.flux_expression * coefficient
                continue

        to_add = []

        ind_var_id = "y_%s" % reaction.id
        aux_var_id = "u_%s" % reaction.id
        try:
            ind_var = model.solver.variables[ind_var_id]
            aux_var = model.solver.variables[aux_var_id]
        except KeyError:
            ind_var = model.solver.interface.Variable(ind_var_id, type='binary')
            aux_var = model.solver.interface.Variable(aux_var_id, lb=0)
            to_add += [ind_var, aux_var]

        aux_variables[reaction.id] = aux_var
        ind_variables[reaction.id] = ind_var

        upper_indicator_constraint_name = "ind_%s_u" % reaction.id
        lower_indicator_constraint_name = "ind_%s_l" % reaction.id

        auxiliary_constraint_a_name = "aux_%s_a" % reaction.id
        auxiliary_constraint_b_name = "aux_%s_b" % reaction.id
        auxiliary_constraint_c_name = "aux_%s_c" % reaction.id
        auxiliary_constraint_d_name = "aux_%s_d" % reaction.id

        try:
            model.solver.constraints[upper_indicator_constraint_name]
        except KeyError:

            # constraint y to be 0 if Sv >= 0 (production)

            #   -M             0                M
            # Sv <-------------|---------------->
            #         y=0      |      y=1

            # -Sv - My <= 0
            # if y = 1 then Sv <= M
            # if y = 0 then Sv >= 0
            upper_indicator_expression = - coefficient * reaction.flux_expression - ind_var * constant
            ind_constraint_u = model.solver.interface.Constraint(upper_indicator_expression,
                                                                 name=upper_indicator_constraint_name,
                                                                 ub=0)

            # -Sv + M(1-y) >= 0
            # if y = 1 then Sv <= 0
            # if y = 0 then Sv <= M
            lower_indicator_expression = - coefficient * reaction.flux_expression + constant - ind_var * constant
            ind_constraint_l = model.solver.interface.Constraint(lower_indicator_expression,
                                                                 name=lower_indicator_constraint_name,
                                                                 lb=0)

            # a) -My + u <= 0
            # b) My + u >= 0

            # if y = 0, u = 0
            # if y = 1, -M <= u <= M
            aux_indicator_expression_a = -constant * ind_var + aux_var
            aux_constraint_a = model.solver.interface.Constraint(aux_indicator_expression_a,
                                                                 name=auxiliary_constraint_a_name,
                                                                 ub=0)

            aux_indicator_expression_b = constant * ind_var + aux_var
            aux_constraint_b = model.solver.interface.Constraint(aux_indicator_expression_b,
                                                                 name=auxiliary_constraint_b_name,
                                                                 lb=0)
            #
            # # c) -M(1-y) + u + viSi <= 0
            # # d) M(1-y) + u + viSi >= 0
            #
            # # if y = 1 then 0 <= u + viSi <= 0
            # # if y = 0 then -M <= u + viSi <= M
            aux_indicator_expression_c = -constant * (1 - ind_var) + aux_var + reaction.flux_expression * coefficient
            aux_constraint_c = model.solver.interface.Constraint(aux_indicator_expression_c,
                                                                 name=auxiliary_constraint_c_name,
                                                                 ub=0)

            aux_indicator_expression_d = constant * (1 - ind_var) + aux_var + reaction.flux_expression * coefficient
            aux_constraint_d = model.solver.interface.Constraint(aux_indicator_expression_d,
                                                                 name=auxiliary_constraint_d_name,
                                                                 lb=0)

            to_add += [ind_constraint_l, ind_constraint_u, aux_constraint_a,
                       aux_constraint_b, aux_constraint_c, aux_constraint_d]

            model.add_cons_vars(to_add)

    max_production_turnover = (1 - fraction) * (turnover / 2)
    # sum(u) <= (1-fraction) * uWT
    constraint_name = "take_more_%s" % metabolite.id
    if constraint_name not in model.constraints:
        decrease_turnover_constraint = model.solver.interface.Constraint(sum(aux_variables.values()),
                                                                         name=constraint_name,
                                                                         ub=max_production_turnover)
        model.add_cons_vars(decrease_turnover_constraint)
    else:
        decrease_turnover_constraint = model.constraints[constraint_name]
        decrease_turnover_constraint.ub = max_production_turnover

    return exchange


def knockout_metabolite(model, metabolite, ignore_transport=True, allow_accumulation=True):
    """
    Inhibits the usage of a metabolite based on a reference flux distributions.

    Parameters
    ----------
    model : Model
        A constraint-based model.
    metabolite: cobra.Metabolite
        A metabolite.
    ignore_transport : bool
        Choose to ignore transport reactions.
    allow_accumulation : bool
        Allow to accumulate the metabolite (add a exchange reaction).

    Returns
    -------
    cameo.core.Reaction, None
        If allow accumulation returns the exchange reaction associated with the metabolite.
    """
    assert isinstance(model, Model)

    reactions = metabolite.reactions

    if ignore_transport:
        reactions = [r for r in reactions if not len(set(m.compartment for m in r.metabolites)) > 1]

    exchanges = model.exchanges

    for reaction in reactions:
        assert isinstance(reaction, Reaction)

        if reaction in exchanges:
            continue

        if reaction.reversibility:
            reaction.bounds = (0, 0)
        elif reaction.metabolites[metabolite] < 0:
            reaction.lower_bound = 0

    exchange = None

    if allow_accumulation:
        species_id = metabolite.id[:-2]
        if "EX_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("EX_%s_e" % species_id)
        elif "DM_%s_e" % species_id in exchanges:
            exchange = model.reactions.get_by_id("DM_%s_e" % species_id)
        else:
            reaction_id = "KO_%s" % metabolite.id
            if reaction_id in model.reactions:
                exchange = model.reactions.get_by_id(reaction_id)
            else:
                exchange = model.add_boundary(metabolite, type="ko sink", reaction_id=reaction_id, lb=0)

    return exchange


def apply_anti_metabolite(model, metabolites, essential_metabolites, reference, inhibition_fraction=.0,
                          competition_fraction=.0, allow_accumulation=True):
    """
    Apply a metabolite in the context of a model without knowing if it is activating or inhibiting.

    Parameters
    ----------
    model : cameo.core.SolverBasedModel
        A constraint-based model.
    metabolites : list
        Metabolites of the same species.
    essential_metabolites : list
        A list of essential metabolites.
    reference : dict, cameo.core.FluxDistributionResult
        A flux distribution.
    inhibition_fraction : float
        How much a metabolite inhibits.
    competition_fraction : float
        How much a metabolite competes.
    allow_accumulation : bool
        Allow accumulation of the metabolite.

    Returns
    -------
    set
        Exchange reactions added for accumulation.

    """
    exchanges = set()

    if any(met in essential_metabolites for met in metabolites):
        for metabolite in metabolites:
            exchanges.add(compete_metabolite(model,
                                             metabolite,
                                             reference,
                                             allow_accumulation=allow_accumulation,
                                             fraction=competition_fraction))
    else:
        for metabolite in metabolites:
            exchanges.add(inhibit_metabolite(model,
                                             metabolite,
                                             reference,
                                             allow_accumulation=allow_accumulation,
                                             fraction=inhibition_fraction))

    return exchanges



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pybel
from pybel import readfile

from numpy import nan

from marsi.config import default_session
from marsi.io.db import Metabolite, Reference, Synonym
from marsi.chemistry import openbabel


def build_database(data, data_dir, with_zinc=True, session=default_session):
    """
    Builds then Molecules database.
    It requires that the input files have been downloaded.

    See Also
    --------
    Initialization documentation.

    Parameters
    ----------
    data : module
        The marsi.io.data module
    data_dir : str
        The path to where data is stored.

    """
    keys = set()
    i = 0
    chebi_structures_file = os.path.join(data_dir, "chebi_lite_3star.sdf")
    i = upload_chebi_entries(chebi_structures_file, data.chebi, i=i, session=session, keys=keys)
    print("Added %i" % i)
    session.commit()
    drugbank_structures_file = os.path.join(data_dir, "drugbank_open_structures.sdf")
    i = upload_drugbank_entries(drugbank_structures_file, data.drugbank, i=i, session=session, keys=keys)
    print("Added %i" % i)
    session.commit()
    kegg_mol_files_dir = os.path.join(data_dir, "kegg_mol_files")
    i = upload_kegg_entries(kegg_mol_files_dir, data.kegg, i=i, session=session, keys=keys)
    print("Added %i" % i)
    session.commit()
    pubchem_sdf_files_dir = os.path.join(data_dir, "pubchem_sdf_files")
    i = upload_pubchem_entries(pubchem_sdf_files_dir, data.pubchem, i=i, session=session, keys=keys)
    print("Added %i" % i)
    session.commit()
    zinc_data_file = os.path.join(data_dir, "zinc_16.sdf.gz")
    if with_zinc:
        i = upload_zinc_entries(zinc_data_file, i=i, session=session, keys=keys)
        print("Added %i" % i)

    session.commit()
    return i


def _add_molecule(mol, synonyms, database, identifier, is_analog, session=default_session, keys=None):
    """
    Add a molecule to the database. It checks for radicals, and it only adds complete molecules.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule parsed by openbabel.
    synonyms : list
        A list of strings with common names for the molecule.
    database : str
        The database from where this molecule was retrieved.
    identifier : str
        The molecule identifier at database.
    is_analog : bool
        If the metabolite was labled as an analog.

    """
    if not openbabel.has_radical(mol):
        inchi_key = openbabel.mol_to_inchi_key(mol)
        if len(inchi_key) > 0 and inchi_key not in keys:
            reference = Reference.add_reference(database, identifier)
            synonyms = list(synonyms)
            clean_synonyms = []
            for synonym in synonyms:
                if isinstance(synonym, str):
                    clean_synonyms.append(Synonym.add_synonym(synonym))

            Metabolite.from_molecule(mol, [reference], clean_synonyms, is_analog, session=session, first_time=True)
            keys.add(inchi_key)


def upload_chebi_entries(chebi_structures_file, chebi_data, i=0, session=default_session, keys=None):
    """
    Import ChEBI data
    """
    for mol in readfile("sdf", chebi_structures_file):
        chebi_id = openbabel.mol_chebi_id(mol)
        chebi_id_int = int(chebi_id.split(":")[1])
        chebi_rows = chebi_data.query('compound_id == @chebi_id_int')
        assert chebi_id == "CHEBI:%i" % chebi_id_int, (chebi_id, "CHEBI:%i" % chebi_id_int)

        if len(chebi_rows) > 0:
            synonyms = list(chebi_rows.name)
            _add_molecule(mol, synonyms, 'chebi', chebi_id, True, session=session, keys=keys)
            i += 1
    return i


def upload_drugbank_entries(drugbank_structures_file, drugbank_data, i=0, session=default_session, keys=None):
    """
    Import DrugBank
    """
    for mol in readfile("sdf", drugbank_structures_file):
        drugbank_id = openbabel.mol_drugbank_id(mol)
        drugbank_rows = drugbank_data.query("id == @drugbank_id")
        if len(drugbank_rows) > 0:
            _add_molecule(mol, [drugbank_rows.iloc[0].synonyms[0]], 'drugbank', drugbank_id, False,
                          session=session, keys=keys)
        i += 1
    return i


def upload_kegg_entries(kegg_mol_files_dir, kegg_data, i=0, session=default_session, keys=None):
    """
    Import KEGG
    """
    for mol_file in os.listdir(kegg_mol_files_dir):
        if mol_file[-4:] == ".mol":
            kegg_id = mol_file[:-4]
            try:
                mol = next(readfile("mol", os.path.join(kegg_mol_files_dir, mol_file)))
            except StopIteration:
                continue
            rows = kegg_data.query("kegg_drug_id == @kegg_id")
            synonyms = set(rows.generic_name.values.tolist() + rows.name.values.tolist())
            if None in synonyms:
                synonyms.remove(None)
            if nan in synonyms:
                synonyms.remove(nan)
            try:
                _add_molecule(mol, synonyms, 'kegg', kegg_id, False, session=session, keys=keys)
                i += 1
            except Exception as e:
                print(synonyms)
                raise e

    return i


def upload_pubchem_entries(pubchem_sdf_files_dir, pubchem_data, i=0, session=default_session, keys=None):
    """
    Import PubChem
    """

    for sdf_file in os.listdir(pubchem_sdf_files_dir):
        if sdf_file[-4:] == ".sdf":
            pubchem_id = sdf_file[:-4]
            try:
                mol = next(readfile("mol", os.path.join(pubchem_sdf_files_dir, sdf_file)))
            except StopIteration:
                continue
            rows = pubchem_data.query("compound_id == @pubchem_id")
            synonyms = set(rows.name.values.tolist() + rows.uipac_name.values.tolist())
            if None in synonyms:
                synonyms.remove(None)

            _add_molecule(mol, synonyms, 'pubchem', pubchem_id, True, session=session, keys=keys)
            i += 1

    return i


def upload_zinc_entries(zinc_data_file, i=0, session=default_session, keys=None):
    """
    Add ZINC
    """
    if os.path.isfile(zinc_data_file):
        zinc = pybel.readfile('sdf', zinc_data_file)
        for j, molecule in enumerate(zinc):
            if not openbabel.has_radical(molecule):
                _add_molecule(molecule, [], 'zinc', molecule.title, False, session=session, keys=keys)
                i += 1

            if j % 20000 == 0:
                session.commit()

    return i



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cameo.core.target import Target
from cameo.flux_analysis.analysis import find_essential_metabolites
from gnomic import Genotype, Feature, Ins, Type, genotype_to_string

from marsi.cobra.flux_analysis.manipulation import knockout_metabolite, apply_anti_metabolite
from marsi.utils import search_metabolites


class AntiMetaboliteManipulationTarget(Target):
    """
    Metabolite target used to manipulate fluxes.
    There are three modes of action:
    1. Inhibition
    2. Competition
    3. Knockout (see Metabolite Knockout Target)

    If the metabolite is essential, the fluxes around the metabolite will be increased for the cells to survive. If the
    metabolite is not essential, the fluxes will be decreased as a consequence of the metabolite presence.

    """

    __gnomic_feature_type__ = "antimetabolite"

    def __init__(self, species_id, fraction=0.5, ignore_transport=True, allow_accumulation=True):
        super(AntiMetaboliteManipulationTarget, self).__init__(species_id)
        self.fraction = fraction
        self.ignore_transport = ignore_transport
        self.allow_accumulation = allow_accumulation

    def get_model_target(self, model):
        """
        Finds the metabolites in the model across compartments.

        Parameters
        ----------
        model: cameo.SolverBasedModel
            A constraint-based model.
        Returns
        -------
        list
            A list of cobra.core.metabolite.Metabolite.
        """
        return search_metabolites(model, self.id)

    def apply(self, model, reference=None):
        essential_metabolites = find_essential_metabolites(model)
        target_metabolites = self.get_model_target(model)

        apply_anti_metabolite(model, target_metabolites, essential_metabolites, reference,
                              allow_accumulation=self.allow_accumulation,
                              competition_fraction=self.fraction,
                              inhibition_fraction=self.fraction)

    def _repr_html_(self):  # pragma: no cover
        return "&#x2623;(%.3f)-%s" % (self.fraction, self.id)

    def __str__(self):
        return genotype_to_string(Genotype([self.to_gnomic()]))

    def __repr__(self):
        return "<AntiMetaboliteManipulation %s (%.3f)>" % (self.id, self.fraction)

    def to_gnomic(self):
        accession = Target.to_gnomic(self)
        feature = Feature(name=self.id, accession=accession, type=Type(self.__gnomic_feature_type__),
                          variant=["value=%.5f" % self.fraction])
        return Ins(feature)


class MetaboliteKnockoutTarget(AntiMetaboliteManipulationTarget):
    def __init__(self, species_id, ignore_transport=True, allow_accumulation=True):
        super(MetaboliteKnockoutTarget, self).__init__(species_id, 1.0, ignore_transport, allow_accumulation)

    def apply(self, model, reference=None):
        for metabolite in self.get_model_target(model):
            knockout_metabolite(model, metabolite, self.ignore_transport, self.allow_accumulation)

    def _repr_html_(self):  # pragma: no cover
        return "&#x2623;%s" % self.id

    def __repr__(self):
        return "<MetaboliteKnockout %s>" % self.id

    def to_gnomic(self):
        accession = Target.to_gnomic(self)
        feature = Feature(name=self.id, accession=accession, type=Type(self.__gnomic_feature_type__))
        return Ins(feature)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import gzip
import logging
import os
import pickle
import re
import shutil
import time

import numpy as np
from IProgress import ProgressBar, Percentage
from cameo import fba
from cameo.flux_analysis.analysis import n_carbon
from cobra.core.reaction import Reaction

from marsi import config

__all__ = ['data_dir', 'log_dir', 'pickle_large', 'unpickle_large', 'frange', 'src_dir', 'internal_data_dir']

data_dir = os.path.join(config.prj_dir, "data")
models_dir = os.path.join(config.prj_dir, "models")
log_dir = os.path.join(config.prj_dir, "log")
src_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)))

internal_data_dir = os.path.join(src_dir, 'io', 'files')

INCHI_KEY_TYPE = np.dtype("a27")

BIOMASS_RE = re.compile("biomass", re.IGNORECASE)

MAX_BYTES = 2 ** 31 - 1

logger = logging.getLogger(__name__)


def pickle_large(obj, file_path, progress=False):
    with open(file_path, 'wb') as model_handler:
        bytes_out = pickle.dumps(obj)
        output_size = len(bytes_out)
        if progress:
            pbar = ProgressBar(maxval=output_size, widgets=["Writing ", Percentage()])
            for idx in pbar(range(0, output_size, MAX_BYTES)):
                model_handler.write(bytes_out[idx:idx + MAX_BYTES])
        else:
            for idx in range(0, output_size, MAX_BYTES):
                model_handler.write(bytes_out[idx:idx + MAX_BYTES])


def unpickle_large(file_path, progress=False):
    input_size = os.path.getsize(file_path)
    logger.debug("Input size: %f bytes" % input_size)
    with open(file_path, 'rb') as file_handler:
        bytes_in = bytearray(0)

        if progress:
            pbar = ProgressBar(maxval=input_size, widgets=["Loading ", Percentage()])
            for _ in pbar(range(0, input_size, MAX_BYTES)):
                bytes_in += file_handler.read(MAX_BYTES)
        else:
            for _ in range(0, input_size, MAX_BYTES):
                bytes_in += file_handler.read(MAX_BYTES)

    return pickle.loads(bytes_in)


def frange(start, stop=None, steps=10):
    """
    Float range generator.

    Generates *steps* equally separated between *start* and *stop*.
    If *stop* is None, the values are between 0 and *start*

    Parameters
    ----------
    start : float
        The initial value.
    stop : float
        The final value.
    steps : int
        Number of values to generate.

    Returns
    -------
    generator
        A generator that yields float.
    """
    if stop is None:
        stop = start
        start = 0

    # Python 2 division of int returns int
    start = float(start)
    stop = float(stop)

    step_size = (stop - start) / float(steps)
    logger.debug("Step size %f" % step_size)
    for i in range(steps):
        logger.debug("Iteration %i: %f" % (i + 1, i * step_size))
        yield start + i * step_size


def unique(l):
    """
    Removes repeated values from a list.

    Parameters
    ----------
    l: list

    Returns
    -------
    list
        The same list with only unique values.
    """
    s = set()
    n = 0
    for x in l:
        if x not in s:
            s.add(x)
            l[n] = x
            n += 1
    del l[n:]


def timing(debug=False):  # pragma: no cover
    def function_wrapper(func):
        if debug:
            def debug_wrap_func(*args, **kwargs):
                start = time.time()
                ret = func(*args, **kwargs)
                stop = time.time()
                if config.log.level >= config.Level.DEBUG:
                    print('%s function took %0.3f ms' % (func.__name__, (stop - start) * 1000.0))
                return ret
            return debug_wrap_func
        else:
            def wrap_func(*args, **kwargs):
                start = time.time()
                ret = func(*args, **kwargs)
                stop = time.time()
                print('%s function took %0.3f ms' % (func.__name__, (stop - start) * 1000.0))
                return ret
            return wrap_func
    return function_wrapper


def default_carbon_sources(model):
    solution = fba(model)
    carbon_sources = []

    for ex in model.exchanges:
        assert isinstance(ex, Reaction)
        if ex.lower_bound < 0 and solution[ex.id] < 0 < n_carbon(ex):
            logger.debug("Found carbon source: %s")
            carbon_sources.append(ex)

    return carbon_sources


def gunzip(file):
    assert file[-3:] == ".gz"
    in_name = file
    out_name = file[0:-3]
    with gzip.open(in_name, 'rb') as f_in, open(out_name, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


def search_metabolites(model, species_id, ignore_external=True):
    if ignore_external:
        return model.metabolites.query(lambda mid: mid[:-2] == species_id and mid[-2:] != "_e", attribute='id')
    else:
        return model.metabolites.query(lambda mid: mid[:-2] == species_id, attribute='id')



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import numpy as np
from cachetools import LRUCache
from scipy.spatial import ConvexHull

from marsi.chemistry.common_ext import tanimoto_coefficient, tanimoto_distance, rmsd, monte_carlo_volume
from scipy.spatial.qhull import QhullError


__all__ = ["rmsd", "tanimoto_coefficient", "tanimoto_distance", "monte_carlo_volume",
           "INCHI_KEY_REGEX", 'SOLUBILITY']


inchi_key_lru_cache = LRUCache(maxsize=512)


SOLUBILITY = {
    "high": lambda sol: sol > 0.00006,
    "medium": lambda sol: 0.00001 <= sol <= 0.00006,
    "low": lambda sol: sol < 0.00001,
    "all": lambda sol: True
}


INCHI_KEY_REGEX = re.compile("[0-9A-Z]{14}\-[0-9A-Z]{8,10}\-[0-9A-Z]")


def convex_hull_volume(xyz):
    try:
        return ConvexHull(xyz).volume
    except (QhullError, ValueError):
        return np.nan



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import six
from bitarray import bitarray
from sqlalchemy import func
from sqlalchemy import inspect

from marsi.chemistry import rdkit

from marsi.chemistry import openbabel

from sqlalchemy import Boolean, Integer, String, Table, Text
from sqlalchemy import Column, ForeignKey, Index, UniqueConstraint
from sqlalchemy import TypeDecorator
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import validates, relationship, backref
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.elements import and_

from marsi.chemistry.common import INCHI_KEY_REGEX
from marsi.config import default_session

__all__ = ['Database', 'Metabolite', 'Reference']


Base = declarative_base()


class ColumnVector(object):
    def __init__(self, collection, session, column):
        assert issubclass(collection, Base)
        self.collection = collection
        self.session = session
        self.column = column

    def apply(self, *args, **kwargs):
        return self.collection.apply(*args, context=self.column, **kwargs)


class CollectionWrapper(object):
    def __init__(self, collection, session=default_session):
        assert issubclass(collection, Base)
        self.collection = collection
        self.session = session
        self.mapper = inspect(self.collection)

    def dump(self, i):
        return [self.__getitem__(item).dump() for item in range(i)]

    def restore(self, dump, session=default_session):
        for _dump in dump:
            self.collection.restore(_dump, session=session)

    def __len__(self):
        return self.session.query(func.count(self.collection.id))

    def __iter__(self):
        return self.session.query(self.collection).yield_per(1000)

    def __getitem__(self, item):
        return self.session.query(self.collection).filter(self.collection.id == item + 1).one()

    def __getattribute__(self, item):
        try:
            return super(CollectionWrapper, self).__getattribute__(item)
        except AttributeError:
            if item in [attr.key for attr in self.mapper.attrs]:
                return ColumnVector(self.collection, self.session, item)
            else:
                return getattr(self.collection, item)


class Fingerprint(TypeDecorator):
    impl = String

    def process_bind_param(self, value, dialect):
        return value.to01()

    def process_result_value(self, value, dialect):
        return bitarray(value)

    def copy(self, **kw):
        return Fingerprint(self.impl.length)


references_table = Table('metabolite_references', Base.metadata,
                         Column('metabolite_id', Integer, ForeignKey('metabolites.id')),
                         Column('reference_id', Integer, ForeignKey('references.id')))

synonyms_table = Table('metabolite_synonyms', Base.metadata,
                       Column('metabolite_id', Integer, ForeignKey('metabolites.id')),
                       Column('synonym_id', Integer, ForeignKey('synonyms.id')))


class Synonym(Base):
    __tablename__ = "synonyms"
    id = Column(Integer, primary_key=True)
    synonym = Column(String(500), nullable=False)

    __table_args__ = (Index('uq_synonyms', synonym, unique=True), )

    @classmethod
    def add_synonym(cls, synonym, session=default_session):
        query = session.query(cls).filter(cls.synonym == synonym)

        try:
            synonym = query.one()
        except NoResultFound:
            synonym = cls(synonym=synonym)
            session.add(synonym)
            session.flush()

        return synonym

    def to_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class Reference(Base):
    __tablename__ = "references"

    id = Column(Integer, primary_key=True)
    database = Column(String(100), nullable=False)
    accession = Column(String(100), nullable=False)

    __table_args__ = (UniqueConstraint('database', 'accession', name='_database_accession_uc'), )

    @classmethod
    def add_reference(cls, database, accession, session=default_session):
        database = database.strip()
        accession = accession.strip()
        query = session.query(cls).filter(and_(cls.database == database, cls.accession == accession))

        try:
            reference = query.one()
        except NoResultFound:
            reference = cls(database=database, accession=accession)
            session.add(reference)
            session.flush()

        return reference

    def __str__(self):
        return "%s: %s" % (self.database, self.accession)

    def to_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class MetaboliteFingerprint(Base):
    __tablename__ = 'metabolite_fingerprints'

    id = Column(Integer, primary_key=True)
    metabolite_id = Column(Integer, ForeignKey('metabolites.id'))
    fingerprint_type = Column(String(10), nullable=False)
    fingerprint = Column(Fingerprint(2048), nullable=False)

    metabolite = relationship("Metabolite", backref=backref(
        "_fingerprints",
        collection_class=attribute_mapped_collection("fingerprint_type"),
        cascade="all, delete-orphan"))

    __table_args__ = (
        UniqueConstraint('metabolite_id', 'fingerprint_type', name='_fp_type_uc'),
    )

    def to_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class Metabolite(Base):
    __tablename__ = "metabolites"

    id = Column(Integer, primary_key=True)
    inchi_key = Column(String(27), nullable=False)
    inchi = Column(String(5000), nullable=False)
    analog = Column(Boolean, default=False)
    formula = Column(String(500), nullable=False)
    num_atoms = Column(Integer, nullable=False)
    num_bonds = Column(Integer, nullable=False)
    num_rings = Column(Integer, nullable=False)
    sdf = Column(Text, nullable=True)

    # solubility = Column(Float)

    references = relationship("Reference", secondary=references_table)
    synonyms = relationship("Synonym", secondary=synonyms_table)
    fingerprints = association_proxy('_fingerprints', 'fingerprint',
                                     creator=lambda key, value: MetaboliteFingerprint(fingerprint_type=key,
                                                                                      fingerprint=value))

    __table_args__ = (
        Index('uq_inchi_key', inchi_key, unique=True),
    )

    @validates('inchi_key')
    def validate_inchi_key(self, key, inchi_key):
        if not INCHI_KEY_REGEX.match(inchi_key):
            raise ValueError("InChI Key %s is not valid" % inchi_key)
        else:
            return inchi_key

    @classmethod
    def get(cls, inchi_key, session=default_session):
        """
        Retrieves a metabolite using the InChI Key.

        Parameters
        ----------
        inchi_key : str
            A valid InChi Key.
        session : sqlalchemy.orm.session.Session
            A database session.

        Returns
        -------
        Metabolite

        Raises
        ------
        KeyError
            If the InChI Key is not available.
        """
        query = session.query(cls).filter(cls.inchi_key == inchi_key)
        try:
            return query.one()
        except NoResultFound:
            raise KeyError(inchi_key)

    @classmethod
    def from_molecule(cls, molecule, references, synonyms, analog=False, session=default_session, first_time=False):
        inchi_key = openbabel.mol_to_inchi_key(molecule)
        try:
            if first_time:
                raise KeyError()

            metabolite = cls.get(inchi_key=inchi_key)
            for reference in references:
                if reference not in metabolite.references:
                    metabolite.references.append(reference)
            for synonym in synonyms:
                if synonym not in metabolite.synonyms:
                    metabolite.synonyms.append(synonym)
        except KeyError:
            metabolite = Metabolite(inchi_key=inchi_key,
                                    inchi=openbabel.mol_to_inchi(molecule),
                                    analog=analog,
                                    formula=molecule.formula,
                                    sdf=openbabel.molecule_to_sdf(molecule),
                                    num_atoms=molecule.OBMol.NumAtoms(),
                                    num_bonds=molecule.OBMol.NumBonds(),
                                    num_rings=len(molecule.OBMol.GetSSSR()))
            for reference in references:
                if reference not in metabolite.references:
                    metabolite.references.append(reference)
            for synonym in synonyms:
                if synonym not in metabolite.synonyms:
                    metabolite.synonyms.append(synonym)

            fingerprint = openbabel.fingerprint(molecule, 'maccs')
            bits = openbabel.fp_bits.get('maccs', 2048)
            metabolite.fingerprints['maccs'] = openbabel.fingerprint_to_bits(fingerprint, bits)

            session.add(metabolite)

        return metabolite

    # NOTE: Hack to get SDF files correct
    @property
    def _sdf(self):
        if self.sdf is None:
            raise ValueError("SDF is not available")
        if self.sdf.startswith("OpenBabel"):
            return "QuickFix1234\n" + self.sdf
        else:
            return self.sdf

    def calc_solubility(self):
        molecule = self.molecule(library='openbabel')
        return openbabel.solubility(molecule, log_value=False) * molecule.molwt

    @property
    def volume(self):
        mol = self.molecule(library='openbabel')
        return openbabel.monte_carlo_volume(mol, self.atom_coordinates, tolerance=1, max_iterations=100)

    def molecule(self, library='openbabel', get3d=True):
        if library == 'openbabel':
            if get3d and self.sdf is not None:
                molecule = openbabel.sdf_to_molecule(self._sdf, from_file=False)
                molecule.title = ""
                return molecule
            else:
                return openbabel.inchi_to_molecule(str(self))
        elif library == 'rdkit':
            try:
                if get3d and self.sdf is not None:
                    return rdkit.sdf_to_molecule(self._sdf, from_file=False)
                else:
                    return rdkit.inchi_to_molecule(str(self))
            except Exception as e:
                print(self.inchi_key)
                raise e
        else:
            raise ValueError("Invalid library: %s, please choose between `openbabel` or `rdkit`")

    def fingerprint(self, fingerprint_format='maccs'):
        if fingerprint_format not in self.fingerprints:
            ob_molecule = self.molecule(get3d=False)
            fingerprint = openbabel.fingerprint(ob_molecule, fingerprint_format)
            bits = openbabel.fp_bits.get(fingerprint_format, 2048)
            self.fingerprints[fingerprint_format] = openbabel.fingerprint_to_bits(fingerprint, bits)

        return self.fingerprints[fingerprint_format]

    def _repr_html_(self):
        mol = self.molecule(library='openbabel')
        mol.make3D(forcefield='mmff94')
        mol.removeh()
        structure = mol._repr_html_() or openbabel.mol_to_svg(mol)
        references = "; ".join(str(r) for r in self.references)
        synonyms = ", ".join([s.synonym for s in self.synonyms])
        return """
<table>
    <tbody>
        <tr><td><strong>Formula</strong></td><td>%s</td></tr>
        <tr><td><strong>InChi</strong></td><td>%s</td></tr>
        <tr><td><strong>InChi Key</strong></td><td>%s</td></tr>
        <tr><td><strong>Structure</strong></td><td>%s</td></tr>
        <tr><td><strong>DBs</strong></td><td>%s</td></tr>
        <tr><td><strong>Synonyms</strong></td><td>%s</td></tr>
    </tbody>
</table>""" % (mol.formula, self.inchi, self.inchi_key, structure, references, synonyms)

    def __repr__(self):
        return "Metabolite %s [%s]" % (self.inchi_key, ", ".join(str(r) for r in self.references))

    def __str__(self):
        """
        Metabolites are represented by it's InChI.

        """
        return self.inchi

    def to_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def dump(self):
        references = [ref.to_dict() for ref in self.references]
        synonyms = [syn.to_dict() for syn in self.synonyms]
        fingerprints = {key: fp.to01() for key, fp in six.iteritems(self.fingerprints)}
        metabolite = self.to_dict()

        return dict(metabolite=metabolite, references=references,
                    synonyms=synonyms, fingerprints=fingerprints)

    @classmethod
    def restore(cls, dump, session=default_session):
        metabolite = cls(**dump['metabolite'])
        references = [Reference.add_reference(ref['database'], ref['accession'], session) for ref in dump['references']]
        synonyms = [Synonym.add_synonym(syn['synonym'], session) for syn in dump['synonyms']]
        metabolite.references = references
        metabolite.synonyms = synonyms
        for key, fingerprint in six.iteritems(dump['fingerprints']):
            metabolite.fingerprints[key] = bitarray(fingerprint)
        session.add(metabolite)
        session.commit()


class Database:
    metabolites = CollectionWrapper(Metabolite)



"""baseline

Revision ID: c1b678211e9d
Revises:
Create Date: 2017-04-12 10:21:55.050636

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'c1b678211e9d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "synonyms",
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('synonym', sa.String(500), nullable=False),
        sa.Index('uq_synonyms', 'synonym', unique=True)
    )

    op.create_table(
        'references',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('database', sa.String(100), nullable=False),
        sa.Column('accession', sa.String(100), nullable=False),
        sa.UniqueConstraint('database', 'accession', name='_database_accession_uc')
    )

    op.create_table(
        'metabolites',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('inchi_key', sa.String(27), nullable=False),
        sa.Column('inchi', sa.String(5000), nullable=False),
        sa.Column('analog', sa.Boolean, default=False),
        sa.Column('formula', sa.String(500), nullable=False),
        sa.Column('num_atoms', sa.Integer, nullable=False),
        sa.Column('num_bonds', sa.Integer, nullable=False),
        sa.Column("num_rings", sa.Integer, nullable=False),
        sa.Column('sdf', sa.Text, nullable=True),
        sa.Index('uq_inchi_key', 'inchi_key', unique=True),
    )

    op.create_table(
        'metabolite_fingerprints',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('metabolite_id', sa.Integer, sa.ForeignKey('metabolites.id')),
        sa.Column('fingerprint_type', sa.String(10), nullable=False),
        sa.Column('fingerprint', sa.String(2048), nullable=False),
        sa.UniqueConstraint('metabolite_id', 'fingerprint_type', name='_fp_type_uc')
    )

    op.create_table(
        'metabolite_references',
        sa.Column('metabolite_id', sa.Integer, sa.ForeignKey('metabolites.id')),
        sa.Column('reference_id', sa.Integer, sa.ForeignKey('references.id'))
    )

    op.create_table(
        'metabolite_synonyms',
        sa.Column('metabolite_id', sa.Integer, sa.ForeignKey('metabolites.id')),
        sa.Column('synonym_id', sa.Integer, sa.ForeignKey('synonyms.id'))
    )


def downgrade():
    op.drop_table("metabolite_fingerprints")
    op.drop_table("metabolite_references")
    op.drop_table("metabolite_synonyms")
    op.drop_table("synonyms")
    op.drop_table("references")
    op.drop_table("metabolites")



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement

import os
import tempfile

from pandas import read_excel

import pubchempy
from bioservices import ChEBI

from IProgress import ProgressBar, Bar, ETA
from alembic import command
from alembic.config import Config
from cement.core.controller import CementBaseController, expose

from marsi.config import db_url
from marsi.io.build_database import build_database
from marsi.io.parsers import parse_chebi_data, parse_pubchem, parse_kegg_brite
from marsi.io.retriaval import retrieve_chebi_names, retrieve_chebi_relation, retrieve_chebi_vertice, \
    retrieve_chebi_structures, retrieve_drugbank_open_structures, retrieve_drugbank_open_vocabulary, \
    retrieve_bigg_reactions, retrieve_bigg_metabolites, retrieve_kegg_brite, retrieve_pubchem_mol_files, \
    retrieve_kegg_mol_files, retrieve_zinc_structures
from marsi.utils import data_dir, src_dir, internal_data_dir

from marsi.io.db import Reference, Synonym, Metabolite
from marsi.io.enrichment import find_best_chebi_structure

from marsi.chemistry import openbabel


PUBCHEM_COMPOUND = "PubChem Compound"
CHEBI = "ChEBI"


class DatabaseController(CementBaseController):
    # TODO: migrate database
    class Meta:
        label = 'db'
        stacked_on = 'base'
        stacked_type = 'nested'
        description = "Initialise MARSI (download data and build initial database)"
        arguments = [
            (['--drugbank-version'], dict(help="DrugBank version (5.0.3)")),
            (['--with-zinc'], dict(help="Include Zinc", action="store_true"))
        ]

    @expose(hide=True)
    def default(self):
        print("MARSI database commands. If it is the first time run 'marsi db init' For details run marsi init --help")

    @expose(hide=False, help="Build and populate database")
    def init(self):
        self.migrate()
        self.download()
        self.build_data()
        self.build_database()

    @expose(hide=True)
    def migrate(self):
        """
        Run database migration.
        """
        alembic_cfg = Config(os.path.join(src_dir, "alembic.ini"))
        alembic_cfg.set_section_option("alembic", "sqlalchemy.url", db_url)
        script_location = alembic_cfg.get_section_option("alembic", "script_location")
        alembic_cfg.set_section_option("alembic", "script_location", os.path.join(src_dir, script_location))
        command.upgrade(alembic_cfg, "head")

    @expose(hide=False, help="Download all files")
    def download(self):
        self.download_chebi()
        self.download_drug_bank()
        self.download_kegg()
        self.download_pubchem()
        self.download_bigg()
        if self.app.pargs.with_zinc:
            self.download_zinc()

    @expose(help="Retrieve ChEBI files (part of download)")
    def download_chebi(self):
        pbar = ProgressBar(maxval=4, widgets=["Downloading ChEBI files", Bar(), ETA()])
        pbar.start()
        retrieve_chebi_names()
        pbar.update(1)
        retrieve_chebi_relation()
        pbar.update(2)
        retrieve_chebi_vertice()
        pbar.update(3)
        retrieve_chebi_structures()
        pbar.update(4)
        pbar.finish()

    @expose(help="Retrieve DrugBank files (part of download)")
    def download_drug_bank(self):
        pbar = ProgressBar(maxval=2, widgets=["Downloading DrugBank files", Bar(), ETA()])
        pbar.start()
        retrieve_drugbank_open_structures(self.app.pargs.drugbank_version or "5.0.3")
        pbar.update(1)
        retrieve_drugbank_open_vocabulary(self.app.pargs.drugbank_version or "5.0.3")
        pbar.update(2)
        pbar.finish()

    @expose(help="Retrieve BIGG files (part of download)")
    def download_bigg(self):
        pbar = ProgressBar(maxval=2, widgets=["Downloading BIGG files", Bar(), ETA()])
        pbar.start()
        retrieve_bigg_reactions()
        pbar.update(1)
        retrieve_bigg_metabolites()
        pbar.update(2)
        pbar.finish()

    @expose(help="Retrieve PubChem files (part of download)")
    def download_pubchem(self):
        pubchem = parse_pubchem(os.path.join(internal_data_dir, "pubchem_compound_analogs_antimetabolites.txt"))
        pubchem_ids = pubchem.compound_id.unique()
        pbar = ProgressBar(maxval=len(pubchem_ids), widgets=["Downloading PubChem files",
                                                                              Bar(), ETA()])
        pbar.start()
        for i in retrieve_pubchem_mol_files(pubchem_ids, dest=data_dir):
            pbar.update(i)
        pbar.finish()

    @expose(help="Retrieve Zinc files (part of download)")
    def download_zinc(self):
        # retrieve_zinc_properties()
        retrieve_zinc_structures()

    @expose(help="Retrieve KEGG files (part of download)")
    def download_kegg(self):
        retrieve_kegg_brite()
        kegg = parse_kegg_brite(os.path.join(data_dir, "kegg_brite_08310.keg"))
        pbar = ProgressBar(maxval=len(kegg.kegg_drug_id.unique()), widgets=["Downloading KEGG files", Bar(), ETA()])
        pbar.start()
        for i in retrieve_kegg_mol_files(kegg, dest=data_dir):
            pbar.update(i)
        pbar.finish()

    @expose(help="Status of init")
    def status(self):
        necessary_files = ["chebi_names_3star.txt", "chebi_vertice_3star.tsv", "chebi_relation_3star.tsv",
                           "chebi_lite_3star.sdf", "kegg_brite_08310.keg", "drugbank_open_vocabulary.csv",
                           "drugbank_open_structures.sdf"]

        if self.app.pargs.with_zinc:
            necessary_files.append("zinc_16.sdf.gz",)

        missing = []

        for file_name in necessary_files:
            file_path = os.path.join(data_dir, file_name)

            if not os.path.isfile(file_path):
                missing.append(file_name)
                print(file_name.ljust(45) + "| FAIL: File does not exist")
            else:
                file_size = os.path.getsize(file_path)

                if file_size == 0:
                    missing.append(file_name)
                    print(file_name.ljust(45) + "| (%.1f MB) FAIL: File is too small" % (file_size / 1048576))
                else:
                    print(file_name.ljust(45) + "| (%.1f MB) OK" % (file_size / 1048576))

        return missing

    @expose(help="Build data (part of download)")
    def build_data(self):
        missing = self.status()
        if len(missing):
            print("Cannot continue without the missing files")
            exit(1)

        print("\nStart building data:")
        print("--------------------------------------------\n")
        print("Building ChEBI:")
        chebi_names_file = os.path.join(data_dir, "chebi_names_3star.txt")
        chebi_vertice_file = os.path.join(data_dir, "chebi_vertice_3star.tsv")
        chebi_relation_file = os.path.join(data_dir, "chebi_relation_3star.tsv")
        chebi_data = parse_chebi_data(chebi_names_file, chebi_vertice_file, chebi_relation_file)
        chebi_data.to_csv(os.path.join(data_dir, "chebi_analogues_filtered.csv"))
        print("Complete!")
        print("--------------------------------------------")
        print("Building PubChem:")
        pubchem = parse_pubchem(os.path.join(internal_data_dir, "pubchem_compound_analogs_antimetabolites.txt"))
        pubchem.to_csv(os.path.join(data_dir, "pubchem_data.csv"))
        print("Complete!")
        print("--------------------------------------------")
        print("Building KEGG:")
        kegg = parse_kegg_brite(os.path.join(data_dir, "kegg_brite_08310.keg"))
        kegg.to_csv(os.path.join(data_dir, "kegg_data.csv"))
        print("Complete!")

    @expose(help="Build database")
    def build_database(self):
        from marsi.io import data
        build_database(data, self.app.pargs.with_zinc, data_dir)

    @expose(help="Add known analogs")
    def add_known_analogs(self):
        chebi_client = ChEBI()
        print("Updating data from 'antimetabolites.xlsx'")

        anti_metabolites = read_excel(os.path.join(internal_data_dir, "antimetabolites.xlsx"), sheetname="Sheet1")
        anti_metabolites = anti_metabolites[anti_metabolites.Identifier != "?"]

        for _, row in anti_metabolites.iterrows():
            if row.Database == PUBCHEM_COMPOUND:
                temp_file = tempfile.mktemp(suffix="sdf", prefix=str(row.Identifier))
                pubchempy.download("SDF", temp_file, row.Identifier, overwrite=True)
                molecule = openbabel.sdf_to_molecule(temp_file)
                reference = Reference.add_reference("pubchem", str(row.Identifier))

            elif row.Database == CHEBI:
                entity = chebi_client.getCompleteEntity(row.Identifier)
                molecule = openbabel.mol_to_molecule(find_best_chebi_structure(entity), False)
                reference = Reference.add_reference("chebi", str(row.Identifier))
            else:
                print('{:16s}\t{:12s}\t{:25s}\t- Skip'.format(row.Database, str(row.Identifier), row.Target))
                continue

            synonym = Synonym.add_synonym(row.Name)
            try:
                Metabolite.get(inchi_key=openbabel.mol_to_inchi_key(molecule))
                print('{:16s}\t{:12s}\t{:25s}\t- OK'.format(row.Database, str(row.Identifier), row.Target))
            except KeyError:
                Metabolite.from_molecule(molecule, [reference], [synonym], analog=True, first_time=False)
                print('{:16s}\t{:12s}\t{:25s}\t- Added'.format(row.Database, str(row.Identifier), row.Target))



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import combinations

from pandas import melt, DataFrame

from bokeh.charts import Bar
from bokeh.io import show


def summary_plot(summary, dbs=['chebi', 'kegg', 'drugbank', 'pubchem']):
    data = DataFrame(summary)
    data['key'] = data.index
    by_database = melt(data, id_vars='key', value_vars=dbs).dropna(subset=['value'])
    bar = Bar(by_database, label='variable', values='key', color='variable', agg='count')
    show(bar)


def venn_plot(summary, dbs=['chebi', 'kegg', 'drugbank', 'pubchem']):
    sets = {db: set(summary[~summary[db].isnull()].index) for db in dbs}
    for db1, db2 in combinations(dbs, 2):
        l_int = len(sets[db1].intersection(sets[db2]))
        print("%s AND %s: %i" % (db1, db2, l_int))

    for db1, db2, db3 in combinations(dbs, 3):
        l_int = len(sets[db1].intersection(sets[db2]).intersection(db3))
        print("%s AND %s AND %s: %i" % (db1, db2, db3, l_int))

    for db1, db2, db3, db4 in combinations(dbs, 4):
        l_int = len(sets[db1].intersection(sets[db2]).intersection(db3).intersection(db4))
        print("%s AND %s AND %s AND %s: %i " % (db1, db2, db3, db4, l_int))



# Copyright 2016 The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

import numpy
import six
from cameo.core.strain_design import StrainDesign
from cameo.core.target import ReactionKnockoutTarget, ReactionModulationTarget
from cameo.flux_analysis.structural import create_stoichiometric_array, find_coupled_reactions_nullspace
from cameo.flux_analysis.structural import nullspace
from cameo.strain_design.heuristic.evolutionary.objective_functions import ObjectiveFunction
from cobra.core.model import Model
from cobra.core.reaction import Reaction
from cobra.exceptions import OptimizationError
from optlang.exceptions import SolverError
from pandas import DataFrame, Series

from marsi.cobra import utils
from marsi.cobra.strain_design.target import AntiMetaboliteManipulationTarget, MetaboliteKnockoutTarget

logger = logging.getLogger(__name__)


def find_anti_metabolite_knockouts(reaction, ref_flux=0, ignore_metabolites=None, ignore_transport=True,
                                   allow_accumulation=True):
    """
    Generates a dictionary {species_id -> MetaboliteKnockoutTarget}.

    Parameters
    ----------
    reaction: cobra.Reaction
        A COBRA reaction
    ref_flux: float
        The flux from the reference state (0 if unknown)
    ignore_metabolites: list
        A list of metabolites that should not be targeted (currency metabolites, etc.)
    ignore_transport: bool
        If False, also knockout the transport reactions.
    allow_accumulation: bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.

    Returns
    -------
    dict

    """
    assert isinstance(reaction, Reaction)
    assert isinstance(ignore_metabolites, (list, set, tuple))

    if ref_flux != 0:
        substrates = [m for m, coefficient in reaction.metabolites.items()
                      if coefficient * ref_flux > 0 and m.id[:-2] not in ignore_metabolites]
    else:
        if reaction.reversibility:
            substrates = [m for m in reaction.metabolites.keys() if m.id[:-2] not in ignore_metabolites]
        else:
            substrates = [m for m, coefficient in reaction.metabolites.items()
                          if coefficient > 0 and m.id[:-2] not in ignore_metabolites]

    species_ids = [m.id[:-2] for m in substrates]
    result = {}
    for species_id in species_ids:
        result[species_id] = MetaboliteKnockoutTarget(species_id, ignore_transport, allow_accumulation)

    return result


def find_anti_metabolite_modulation(reaction, fold_change, essential_metabolites, ref_flux=0, ignore_metabolites=None,
                                    ignore_transport=True, allow_accumulation=True):
    """
    Generates a dictionary {species_id -> AntiMetaboliteManipulationTarget}.

    If the fold change > 0:
        1. Search for metabolites that are essential.
        2. Calculate fraction using a link function.

    If fold change < 0:
        1. Search for metabolites that are not essential
        2. Calculated fraction is 1 - fold change

    Parameters
    ----------
    reaction : cobra.Reaction
        A COBRA reaction.
    fold_change : float
        The fold change of the reaction flux.
    essential_metabolites : list
        A list of essential metabolites.
    ref_flux : float
        The flux from the reference state (0 if unknown)
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc.).
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.

    Returns
    -------
    dict
        {MetaboliteID -> AntiMetaboliteManipulationTarget}
    """

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if essential_metabolites is None and reaction.model is not None:
        essential_metabolites = utils.essential_species_ids(reaction.model)

    assert isinstance(reaction, Reaction)
    assert isinstance(ignore_metabolites, (list, set, tuple))
    assert isinstance(essential_metabolites, (list, set, tuple))

    if fold_change > 0:
        ignore_metabolites = set(ignore_metabolites) | set(essential_metabolites)

    if ref_flux != 0:
        substrates = [m for m, coefficient in reaction.metabolites.items()
                      if coefficient * ref_flux > 0 and m.id[:-2] not in ignore_metabolites]
    else:
        if reaction.reversibility:
            substrates = [m for m in reaction.metabolites.keys() if m.id[:-2] not in ignore_metabolites]
        else:
            substrates = [m for m, coefficient in reaction.metabolites.items()
                          if coefficient > 0 and m.id[:-2] not in ignore_metabolites]

    species_ids = [m.id[:-2] for m in substrates]
    result = {}

    # Use a link function to convert fold change into ]0, 1]
    if fold_change > 0:
        fraction = 1 / (1 + numpy.exp(-0.5 * fold_change - 5))
    else:
        fraction = 1 - fold_change

    for species_id in species_ids:
        result[species_id] = AntiMetaboliteManipulationTarget(species_id, fraction=fraction,
                                                              ignore_transport=ignore_transport,
                                                              allow_accumulation=allow_accumulation)

    return result


def convert_strain_design_results(model, results, objective_function, simulation_method, simulation_kwargs=None,
                                  ignore_metabolites=None, ignore_transport=True, allow_accumulation=True,
                                  nullspace_matrix=None, essential_metabolites=None, max_loss=0.2):
    """
    Converts a StrainDesignMethodResult into a DataFrame of possible substitutions.

    Parameters
    ----------
    model : cobra.Model
        A COBRA model.
    results : cameo.core.strain_design.StrainDesignMethodResult
        The results of a strain design method.
    objective_function : cameo.strain_design.heuristic.evolutionary.objective_functions.ObjectiveFunction
        The cellular objective to evaluate.
    simulation_method : cameo.flux_analysis.simulation.fba or equivalent
        The method to compute a flux distribution using a COBRA model.
    simulation_kwargs : dict
        The arguments for the simulation_method.
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc).
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    nullspace_matrix : numpy.ndarray
        The nullspace of the model.
    essential_metabolites : list
        A list of essential metabolites.
    max_loss : float
        A number between 0 and 1 for how much the fitness is allowed to drop with the metabolite target.

    Returns
    -------
    pandas.DataFrame
        A data frame with the possible replacements.
    """

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if nullspace_matrix is None:
        nullspace_matrix = nullspace(create_stoichiometric_array(model))

    if simulation_kwargs is None:
        simulation_kwargs = {}

    replacements = DataFrame()
    for index, design in enumerate(results):
        with model:
            design.apply(model)
            solution = simulation_method(model, **simulation_kwargs)
            fitness = objective_function(model, solution, design.targets)
        if fitness <= 0:
            continue

        res = convert_design(model, design, fitness, objective_function, simulation_method,
                             simulation_kwargs=simulation_kwargs, ignore_metabolites=ignore_metabolites,
                             ignore_transport=ignore_transport, allow_accumulation=allow_accumulation,
                             nullspace_matrix=nullspace_matrix, essential_metabolites=essential_metabolites,
                             max_loss=max_loss)

        res['index'] = index
        replacements = replacements.append(res, ignore_index=True)

    return replacements


def convert_design(model, strain_design, fitness, objective_function, simulation_method, simulation_kwargs=None,
                   ignore_metabolites=None, ignore_transport=True, allow_accumulation=True, nullspace_matrix=None,
                   essential_metabolites=None, max_loss=0.2):
    """
    Converts a StrainDesign into a DataFrame of possible substitutions.

    Parameters
    ----------
    model: cobra.Model
        A COBRA model.
    strain_design : cameo.core.strain_design.StrainDesign
        The results of a strain design method.
    objective_function : cameo.strain_design.heuristic.evolutionary.objective_functions.ObjectiveFunction
        The cellular objective to evaluate.
    simulation_method : cameo.flux_analysis.simulation.fba or equivalent
        The method to compute a flux distribution using a COBRA model.
    simulation_kwargs : dict
        The arguments for the simulation_method.
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc).
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    nullspace_matrix : numpy.ndarray
        The nullspace of the model.
    essential_metabolites : list
        A list of essential metabolites.
    max_loss : float
        A number between 0 and 1 for how much the fitness is allowed to drop with the metabolite target.

    Returns
    -------
    pandas.DataFrame
        A data frame with the possible replacements.
    """

    if simulation_kwargs is None:
        simulation_kwargs = {}

    if simulation_kwargs.get('reference', None) is None:
        reference = {}
    else:
        reference = simulation_kwargs['reference']

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if nullspace_matrix is None:
        nullspace_matrix = nullspace(create_stoichiometric_array(model))

    coupled_reactions = find_coupled_reactions_nullspace(model, ns=nullspace_matrix)
    coupled_reactions = [{r.id: c for r, c in six.iteritems(g)} for g in coupled_reactions]

    assert isinstance(model, Model)
    assert isinstance(objective_function, ObjectiveFunction)
    assert isinstance(nullspace_matrix, numpy.ndarray)

    def valid_loss(val, base):
        fitness_loss = fitness - val

        return fitness_loss / fitness < max_loss and val - base > 1e-6

    testable_targets = [t for t in strain_design.targets
                        if isinstance(t, (ReactionKnockoutTarget, ReactionModulationTarget))]

    non_testable_targets = {t for t in strain_design.targets
                            if not isinstance(t, (ReactionKnockoutTarget, ReactionModulationTarget))}
    coupled_targets = {}
    for t in testable_targets:
        group = next((g for g in coupled_reactions if t in g), None)
        if group:
            group = frozenset(group)
            if group not in coupled_targets:
                coupled_targets[group] = []
            coupled_targets[group].append(t)

    coupled_targets = [frozenset(t) for k, t in six.iteritems(coupled_targets) if len(t) > 1]
    non_coupled_targets = {t for t in testable_targets if not any(t in group for group in coupled_targets)}

    anti_metabolites = DataFrame(columns=['base_design', 'replaced_target', 'metabolite_targets',
                                          'old_fitness', 'fitness', 'delta'])

    logger.debug("Coupled groups: %i; non-coupled targets %i" % (len(coupled_targets), len(non_coupled_targets)))

    with model as base_model:
        for target in non_testable_targets:
            target.apply(model)

        # test non-coupled targets as before
        for target in non_coupled_targets:
            with base_model:
                for other_target in non_coupled_targets:
                    if other_target != target:
                        other_target.apply(base_model)

                anti_metabolite_targets = convert_target(base_model, target, essential_metabolites,
                                                         ignore_metabolites=ignore_metabolites,
                                                         ignore_transport=ignore_transport,
                                                         allow_accumulation=allow_accumulation,
                                                         reference=reference)

                base_solution = simulation_method(base_model, **simulation_kwargs)
                base_fitness = objective_function(base_model, base_solution, strain_design.targets)

                test_target_substitutions(base_model, strain_design.targets, target, anti_metabolite_targets,
                                          objective_function, fitness, base_fitness, simulation_method,
                                          simulation_kwargs, reference, valid_loss, anti_metabolites)

        # test coupled targets as whole
        for target_group in coupled_targets:
            with base_model:

                for other_target_group in non_coupled_targets:
                    if other_target_group != target_group:
                        for other_target in other_target_group:
                            other_target.apply(model)

                anti_metabolite_targets = convert_target_group(base_model, target_group, essential_metabolites,
                                                               ignore_transport=ignore_transport,
                                                               allow_accumulation=allow_accumulation,
                                                               reference=reference)

                base_solution = simulation_method(model, **simulation_kwargs)
                base_fitness = objective_function(model, base_solution, strain_design.targets)

                test_target_substitutions(base_model, strain_design.targets, target_group, anti_metabolite_targets,
                                          objective_function, fitness, base_fitness, simulation_method,
                                          simulation_kwargs, reference, valid_loss, anti_metabolites)

        return anti_metabolites


def test_target_substitutions(model, all_targets, target, replacement_targets, objective_function, fitness,
                              base_fitness, simulation_method, simulation_kwargs, reference, loss_validation, results):
    fitness2targets = {}
    index = len(results) - 1
    for species_id, replacement_target in replacement_targets.items():
        assert isinstance(replacement_target, AntiMetaboliteManipulationTarget)
        with model:
            try:
                replacement_target.apply(model, reference=reference)
                new_solution = simulation_method(model, **simulation_kwargs)
                new_fitness = objective_function(model, new_solution, all_targets)
                logger.debug("New fitness %s" % new_fitness)
                logger.debug("Solver objective value %s" % new_solution.objective_value)
                for r in objective_function.reactions:
                    logger.debug("%s: %f" % (r, new_solution[r]))
            except OptimizationError:
                logger.debug("Cannot solve %s" % species_id)
                new_fitness = 0
            finally:
                if new_fitness not in fitness2targets:
                    fitness2targets[new_fitness] = []
                fitness2targets[new_fitness].append(replacement_target)
                try:
                    logger.debug("Applying %s yields %f (loss: %f)" %
                                 (species_id, new_fitness, (new_fitness - fitness) / new_fitness))
                except ZeroDivisionError:
                    logger.debug("Applying %s yields %f (loss: %f)" % (species_id, new_fitness, 1))

        # keep only targets that keep the fitness above the valid loss regarding the original fitness.
        fitness2targets = {fit: anti_mets for fit, anti_mets in fitness2targets.items()
                           if loss_validation(fit, base_fitness)}

        if len(fitness2targets) == 0:
            logger.debug("Return target %s, no replacement found" % target)
        else:
            for fit, anti_mets in fitness2targets.items():
                delta = fitness - fit
                results.loc[index] = [StrainDesign(all_targets), target, anti_mets, fitness, fit, delta]
                index += 1


def replace_strain_design_results(model, results, objective_function, simulation_method, simulation_kwargs=None,
                                  ignore_metabolites=None, ignore_transport=True, allow_accumulation=True,
                                  essential_metabolites=None, max_loss=0.2):
    """
    Converts a StrainDesignMethodResult into a DataFrame of possible substitutions.

    Parameters
    ----------
    model : cobra.Model
        A COBRA model.
    results : cameo.core.strain_design.StrainDesignMethodResult
        The results of a strain design method.
    objective_function : cameo.strain_design.heuristic.evolutionary.objective_functions.ObjectiveFunction
        The cellular objective to evaluate.
    simulation_method : cameo.flux_analysis.simulation.fba or equivalent
        The method to compute a flux distribution using a COBRA model.
    simulation_kwargs : dict
        The arguments for the simulation_method.
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc).
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    essential_metabolites : list
        A list of essential metabolites.
    max_loss : float
        A number between 0 and 1 for how much the fitness is allowed to drop with the metabolite target.

    Returns
    -------
    pandas.DataFrame
        A data frame with the possible replacements.
    """

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if simulation_kwargs is None:
        simulation_kwargs = {}

    replacements = DataFrame()
    for index, design in enumerate(results):
        with model:
            design.apply(model)
            solution = simulation_method(model, **simulation_kwargs)
            fitness = objective_function(model, solution, design.targets)
        if fitness <= 0:
            continue

        res = replace_design(model, design, fitness, objective_function, simulation_method,
                             simulation_kwargs=simulation_kwargs, ignore_metabolites=ignore_metabolites,
                             ignore_transport=ignore_transport, allow_accumulation=allow_accumulation,
                             essential_metabolites=essential_metabolites, max_loss=max_loss)

        res['index'] = index
        replacements = replacements.append(res, ignore_index=True)

    return replacements


def replace_design(model, strain_design, fitness, objective_function, simulation_method, simulation_kwargs=None,
                   ignore_metabolites=None, ignore_transport=True, allow_accumulation=True,
                   essential_metabolites=None, max_loss=0.2):
    """
    Converts a StrainDesign into a DataFrame of possible substitutions.

    Parameters
    ----------
    model: cobra.Model
        A COBRA model.
    strain_design : cameo.core.strain_design.StrainDesign
        The results of a strain design method.
    objective_function : cameo.strain_design.heuristic.evolutionary.objective_functions.ObjectiveFunction
        The cellular objective to evaluate.
    simulation_method : cameo.flux_analysis.simulation.fba or equivalent
        The method to compute a flux distribution using a COBRA model.
    simulation_kwargs : dict
        The arguments for the simulation_method.
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc).
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    essential_metabolites : list
        A list of essential metabolites.
    max_loss : float
        A number between 0 and 1 for how much the fitness is allowed to drop with the metabolite target.

    Returns
    -------
    pandas.DataFrame
        A data frame with the possible replacements.
    """

    if simulation_kwargs is None:
        simulation_kwargs = {}

    if simulation_kwargs.get('reference', None) is None:
        reference = {}
    else:
        reference = simulation_kwargs['reference']

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    assert isinstance(model, Model)
    assert isinstance(objective_function, ObjectiveFunction)

    def valid_loss(val, base):
        fitness_loss = fitness - val

        return fitness_loss / fitness < max_loss and val - base > 1e-6

    # Keep track of which targets where tested
    target_test_count = {test.id: 0 for test in strain_design.targets if isinstance(test, ReactionModulationTarget)}
    test_targets = [t for t in strain_design.targets if isinstance(t, ReactionModulationTarget)]
    keep_targets = [t for t in strain_design.targets if not isinstance(t, ReactionModulationTarget)]
    anti_metabolites = DataFrame(columns=['base_design', 'replaced_target', 'metabolite_targets',
                                          'old_fitness', 'fitness', 'delta'])

    def termination_criteria():
        logger.debug("Targets: %i/%i" % (sum(target_test_count.values()), len(test_targets)))
        logger.debug("Anti metabolites: %s" % str(anti_metabolites))
        return len(test_targets) == 0 or all(count == 1 for count in target_test_count.values())

    # Stop when all targets have been replaced or tested more then once.
    while not termination_criteria():
        with model as base_model:
            test_target = test_targets.pop(0)
            target_test_count[test_target.id] += 1

            logger.debug("Testing target %s" % test_target)
            assert test_target not in test_targets

            all_targets = test_targets + keep_targets

            for target in all_targets:
                target.apply(model)

            base_solution = simulation_method(base_model, **simulation_kwargs)
            base_fitness = objective_function(base_model, base_solution, test_targets)

            try:
                anti_metabolite_targets = convert_target(base_model, test_target, essential_metabolites,
                                                         ignore_transport=ignore_transport,
                                                         ignore_metabolites=ignore_metabolites,
                                                         allow_accumulation=allow_accumulation,
                                                         reference=reference)
                test_target_substitutions(base_model, all_targets, test_target, anti_metabolite_targets,
                                          objective_function, fitness, base_fitness, simulation_method,
                                          simulation_kwargs, reference, valid_loss, anti_metabolites)
            except (ValueError, KeyError, SolverError) as e:
                logger.error(str(e))
                continue
            finally:  # put the target back on the list.
                test_targets.append(test_target)

    return anti_metabolites


def convert_target(model, target, essential_metabolites, ignore_transport=True,
                   ignore_metabolites=None, allow_accumulation=True, reference=None):
    """
    Generates a dictionary {species_id -> MetaboliteKnockoutTarget}.

    Parameters
    ----------
    model : cobra.Model
        A COBRA model
    target : ReactionModulationTarget, ReactionKnockoutTarget
        The flux from the reference state (0 if unknown)
    ignore_metabolites : list
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc.)
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    reference : dict
        A dictionary containing the flux values of a reference flux distribution.
    essential_metabolites : list
        A list of essential metabolites

    Returns
    -------
    dict
    """

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if reference is None:
        reference = {}
    elif isinstance(reference, Series):
        reference = reference.to_dict()

    if isinstance(target, ReactionKnockoutTarget):

        substitutions = find_anti_metabolite_knockouts(target.get_model_target(model),
                                                       ref_flux=reference.get(target.id, 0),
                                                       ignore_transport=ignore_transport,
                                                       ignore_metabolites=ignore_metabolites | essential_metabolites,
                                                       allow_accumulation=allow_accumulation)

    elif isinstance(target, ReactionModulationTarget):
        substitutions = find_anti_metabolite_modulation(target.get_model_target(model),
                                                        target.fold_change,
                                                        essential_metabolites,
                                                        ref_flux=reference.get(target.id, 0),
                                                        ignore_transport=ignore_transport,
                                                        ignore_metabolites=ignore_metabolites,
                                                        allow_accumulation=allow_accumulation)

    else:
        raise ValueError("Don't know what to do with the target of type %s" % type(target))

    return substitutions


def convert_target_group(model, target_group, essential_metabolites, ignore_transport=True,
                         ignore_metabolites=None, allow_accumulation=True, reference=None):
    """
    Generates a dictionary {species_id -> MetaboliteKnockoutTarget}.

    Parameters
    ----------
    model : cobra.Model
        A COBRA model
    target_group : dict
        A dictionary of (target: relative coefficient)
    ignore_metabolites : set
        A list of metabolites that should not be targeted (essential metabolites, currency metabolites, etc.)
    ignore_transport : bool
        If False, also knockout the transport reactions.
    allow_accumulation : bool
        If True, create an exchange reaction (unless already exists) to simulate accumulation of the metabolites.
    reference : dict
        A dictionary containing the flux values of a reference flux distribution.
    essential_metabolites : set
        A list of essential metabolites

    Returns
    -------
    tuple
        A dictionary with the antimetabolites to test and another dictionary with the reaction where those metabolites
        belong
    """

    if ignore_metabolites is None:
        ignore_metabolites = set(utils.CURRENCY_METABOLITES)

    if essential_metabolites is None:
        essential_metabolites = utils.essential_species_ids(model)

    if reference is None:
        reference = {}
    elif isinstance(reference, Series):
        reference = reference.to_dict()

    substitutions = {}
    metabolites_targets = {}
    for target in target_group:
        if isinstance(target, ReactionKnockoutTarget):
            ignore = ignore_metabolites | essential_metabolites
            _substitutions = find_anti_metabolite_knockouts(target.get_model_target(model),
                                                            ref_flux=reference.get(target.id, 0),
                                                            ignore_transport=ignore_transport,
                                                            ignore_metabolites=ignore,
                                                            allow_accumulation=allow_accumulation)

        elif isinstance(target, ReactionModulationTarget):
            _substitutions = find_anti_metabolite_modulation(target.get_model_target(model),
                                                             target.fold_change,
                                                             essential_metabolites,
                                                             ref_flux=reference.get(target.id, 0),
                                                             ignore_transport=ignore_transport,
                                                             ignore_metabolites=ignore_metabolites,
                                                             allow_accumulation=allow_accumulation)
        else:
            _substitutions = {}

        substitutions.update(_substitutions)
        for substitution in _substitutions:
            if substitution not in metabolites_targets:
                metabolites_targets[substitution] = set()
            metabolites_targets[substitution].append(target)

    return substitutions, metabolites_targets



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import zipfile
from ftplib import FTP
from io import BytesIO

import bioservices
import pubchempy as pbc
import requests
from IProgress import ProgressBar, Bar, ETA
from six.moves.urllib.request import urlretrieve

from marsi.utils import data_dir, gunzip

BIGG_BASE_URL = "http://bigg.ucsd.edu/static/namespace/"
DRUGBANK_BASE_URL = "https://www.drugbank.ca/"
CHEBI_FTP_URL = "ftp.ebi.ac.uk"
KEGG_BASE_URL = "http://www.genome.jp"
ZINC_BASE_URL = "http://zinc.docking.org/"

ZINC_STRUCTURES = ["16_p0.0.sdf.gz", "16_p0.1.sdf.gz", "16_p0.10.sdf.gz", "16_p0.100.sdf.gz", "16_p0.101.sdf.gz",
                   "16_p0.102.sdf.gz", "16_p0.103.sdf.gz", "16_p0.104.sdf.gz", "16_p0.105.sdf.gz", "16_p0.106.sdf.gz",
                   "16_p0.107.sdf.gz", "16_p0.108.sdf.gz", "16_p0.109.sdf.gz", "16_p0.11.sdf.gz", "16_p0.110.sdf.gz",
                   "16_p0.111.sdf.gz", "16_p0.112.sdf.gz", "16_p0.113.sdf.gz", "16_p0.114.sdf.gz", "16_p0.115.sdf.gz",
                   "16_p0.116.sdf.gz", "16_p0.117.sdf.gz", "16_p0.118.sdf.gz", "16_p0.119.sdf.gz", "16_p0.12.sdf.gz",
                   "16_p0.120.sdf.gz", "16_p0.121.sdf.gz", "16_p0.122.sdf.gz", "16_p0.123.sdf.gz", "16_p0.124.sdf.gz",
                   "16_p0.125.sdf.gz", "16_p0.126.sdf.gz", "16_p0.13.sdf.gz", "16_p0.14.sdf.gz", "16_p0.15.sdf.gz",
                   "16_p0.16.sdf.gz", "16_p0.17.sdf.gz", "16_p0.18.sdf.gz", "16_p0.19.sdf.gz", "16_p0.2.sdf.gz",
                   "16_p0.20.sdf.gz", "16_p0.21.sdf.gz", "16_p0.22.sdf.gz", "16_p0.23.sdf.gz", "16_p0.24.sdf.gz",
                   "16_p0.25.sdf.gz", "16_p0.26.sdf.gz", "16_p0.27.sdf.gz", "16_p0.28.sdf.gz", "16_p0.29.sdf.gz",
                   "16_p0.3.sdf.gz", "16_p0.30.sdf.gz", "16_p0.31.sdf.gz", "16_p0.32.sdf.gz", "16_p0.33.sdf.gz",
                   "16_p0.34.sdf.gz", "16_p0.35.sdf.gz", "16_p0.36.sdf.gz", "16_p0.37.sdf.gz", "16_p0.38.sdf.gz",
                   "16_p0.39.sdf.gz", "16_p0.4.sdf.gz", "16_p0.40.sdf.gz", "16_p0.41.sdf.gz", "16_p0.42.sdf.gz",
                   "16_p0.43.sdf.gz", "16_p0.44.sdf.gz", "16_p0.45.sdf.gz", "16_p0.46.sdf.gz", "16_p0.47.sdf.gz",
                   "16_p0.48.sdf.gz", "16_p0.49.sdf.gz", "16_p0.5.sdf.gz", "16_p0.50.sdf.gz", "16_p0.51.sdf.gz",
                   "16_p0.52.sdf.gz", "16_p0.53.sdf.gz", "16_p0.54.sdf.gz", "16_p0.55.sdf.gz", "16_p0.56.sdf.gz",
                   "16_p0.57.sdf.gz", "16_p0.58.sdf.gz", "16_p0.59.sdf.gz", "16_p0.6.sdf.gz", "16_p0.60.sdf.gz",
                   "16_p0.61.sdf.gz", "16_p0.62.sdf.gz", "16_p0.63.sdf.gz", "16_p0.64.sdf.gz", "16_p0.65.sdf.gz",
                   "16_p0.66.sdf.gz", "16_p0.67.sdf.gz", "16_p0.68.sdf.gz", "16_p0.69.sdf.gz", "16_p0.7.sdf.gz",
                   "16_p0.70.sdf.gz", "16_p0.71.sdf.gz", "16_p0.72.sdf.gz", "16_p0.73.sdf.gz", "16_p0.74.sdf.gz",
                   "16_p0.75.sdf.gz", "16_p0.76.sdf.gz", "16_p0.77.sdf.gz", "16_p0.78.sdf.gz", "16_p0.79.sdf.gz",
                   "16_p0.8.sdf.gz", "16_p0.80.sdf.gz", "16_p0.81.sdf.gz", "16_p0.82.sdf.gz", "16_p0.83.sdf.gz",
                   "16_p0.84.sdf.gz", "16_p0.85.sdf.gz", "16_p0.86.sdf.gz", "16_p0.87.sdf.gz", "16_p0.88.sdf.gz",
                   "16_p0.89.sdf.gz", "16_p0.9.sdf.gz", "16_p0.90.sdf.gz", "16_p0.91.sdf.gz", "16_p0.92.sdf.gz",
                   "16_p0.93.sdf.gz", "16_p0.94.sdf.gz", "16_p0.95.sdf.gz", "16_p0.96.sdf.gz", "16_p0.97.sdf.gz",
                   "16_p0.98.sdf.gz", "16_p0.99.sdf.gz", "16_p1.0.sdf.gz", "16_p1.1.sdf.gz", "16_p1.10.sdf.gz",
                   "16_p1.11.sdf.gz", "16_p1.12.sdf.gz", "16_p1.13.sdf.gz", "16_p1.14.sdf.gz", "16_p1.15.sdf.gz",
                   "16_p1.16.sdf.gz", "16_p1.17.sdf.gz", "16_p1.18.sdf.gz", "16_p1.19.sdf.gz", "16_p1.2.sdf.gz",
                   "16_p1.20.sdf.gz", "16_p1.21.sdf.gz", "16_p1.22.sdf.gz", "16_p1.3.sdf.gz", "16_p1.4.sdf.gz",
                   "16_p1.5.sdf.gz", "16_p1.6.sdf.gz", "16_p1.7.sdf.gz", "16_p1.8.sdf.gz", "16_p1.9.sdf.gz"]

ZINC_SUBSET_16_BASE = "http://zinc.docking.org/db/bysubset/16"


def retrieve_bigg_reactions(dest=os.path.join(data_dir, "bigg_models_reactions.txt")):
    """
    Retrieves bigg reactions file
    """
    bigg_reactions_file = "bigg_models_reactions.txt"
    urlretrieve(BIGG_BASE_URL + bigg_reactions_file, dest)


def retrieve_bigg_metabolites(dest=os.path.join(data_dir, "bigg_models_metabolites.txt")):
    """
    Retrieves bigg metabolites file
    """
    bigg_metabolites_file = "bigg_models_metabolites.txt"
    urlretrieve(BIGG_BASE_URL + bigg_metabolites_file, dest)


def retrieve_drugbank_open_structures(db_version="5.0.3", dest=os.path.join(data_dir, "drugbank_open_structures.sdf")):
    """
    Retrieves Drugbank Open Structures.

    Parameters
    ----------
    db_version: str
        The version of drugbank to retrieve
    """

    encoded_version = db_version.replace(".", "-")
    response = requests.get(DRUGBANK_BASE_URL + "releases/%s/downloads/all-open-structures" % encoded_version)
    with zipfile.ZipFile(BytesIO(response.content)) as zip_ref:
        zip_ref.extractall(data_dir)

    os.rename(os.path.join(data_dir, "open structures.sdf"), dest)


def retrieve_drugbank_open_vocabulary(db_version="5.0.3", dest=os.path.join(data_dir, "drugbank_open_vocabulary.csv")):
    """
    Retrieves Drugbank Open Vocabulary.

    Parameters
    ----------
    db_version: str
        The version of drugbank to retrieve
    """

    encoded_version = db_version.replace(".", "-")
    response = requests.get(DRUGBANK_BASE_URL + "releases/%s/downloads/all-drugbank-vocabulary" % encoded_version)
    with zipfile.ZipFile(BytesIO(response.content)) as zip_ref:
        zip_ref.extractall(data_dir)

    os.rename(os.path.join(data_dir, "drugbank vocabulary.csv"), dest)


def retrieve_chebi_structures(dest=os.path.join(data_dir, "chebi_lite_3star.sdf")):
    """
    Retrieves ChEBI sdf (lite version).
    """
    sdf_file = "ChEBI_lite_3star.sdf.gz"
    chebi_structures_file = dest + ".gz"
    ftp = FTP(CHEBI_FTP_URL)
    ftp.login()
    ftp.cwd('pub/databases/chebi/SDF')
    with open(chebi_structures_file, "wb") as structures_file:
        ftp.retrbinary("RETR %s" % sdf_file, structures_file.write)
    ftp.quit()
    gunzip(chebi_structures_file)


def retrieve_chebi_names(dest=os.path.join(data_dir, "chebi_names_3star.txt")):
    """
    Retrieves ChEBI names.
    """
    gz_file = "names_3star.tsv.gz"
    chebi_names_file = dest + ".gz"
    ftp = FTP(CHEBI_FTP_URL)
    ftp.login()
    ftp.cwd('pub/databases/chebi/Flat_file_tab_delimited')
    with open(chebi_names_file, "wb") as names_file:
        ftp.retrbinary("RETR %s" % gz_file, names_file.write)
    ftp.quit()
    gunzip(chebi_names_file)


def retrieve_chebi_relation(dest=os.path.join(data_dir, "chebi_relation_3star.tsv")):
    """
    Retrieves ChEBI relation data.
    """
    tsv_file = "relation_3star.tsv"
    ftp = FTP(CHEBI_FTP_URL)
    ftp.login()
    ftp.cwd('pub/databases/chebi/Flat_file_tab_delimited')
    with open(dest, "wb") as relation_file:
        ftp.retrbinary("RETR %s" % tsv_file, relation_file.write)
    ftp.quit()


def retrieve_chebi_vertice(dest=os.path.join(data_dir, "chebi_vertice_3star.tsv")):
    """
    Retrieves ChEBI vertice data.
    """
    tsv_file = "vertice_3star.tsv"
    ftp = FTP(CHEBI_FTP_URL)
    ftp.login()
    ftp.cwd('pub/databases/chebi/Flat_file_tab_delimited')
    with open(dest, "wb") as verice_file:
        ftp.retrbinary("RETR %s" % tsv_file, verice_file.write)
    ftp.quit()


def retrieve_kegg_brite(dest=os.path.join(data_dir, "kegg_brite_08310.keg")):
    """
    Retrieves KEGG Brite 08310 (Target-based Classification of Drugs)

    """
    urlretrieve(KEGG_BASE_URL + "/kegg-bin/download_htext?htext=br08310.keg&format=htext&filedir=", dest)


def retrieve_pubchem_mol_files(pubchem_ids, dest=data_dir):
    """
    Retrieves SDF Files from PubChem.
    """
    pubchem_files_path = os.path.join(dest, 'pubchem_sdf_files')

    if not os.path.isdir(pubchem_files_path):
        os.mkdir(pubchem_files_path)

    for i, pubchem_id in enumerate(pubchem_ids):
        try:
            pbc.download('sdf', os.path.join(pubchem_files_path, '%i.sdf' % pubchem_id), int(pubchem_id))
        except IOError:
            # File already exists
            continue
        yield i


def retrieve_kegg_mol_files(kegg, dest=data_dir):
    """
    Retrieves KEGG MOL Files using KEGG REST API.
    """
    kegg_client = bioservices.kegg.KEGG()
    drug_ids = kegg.kegg_drug_id.unique()

    not_found = []

    kegg_mol_files_dir = os.path.join(dest, "kegg_mol_files")
    if not os.path.isdir(kegg_mol_files_dir):
        os.mkdir(kegg_mol_files_dir)

    for i, drug_id in enumerate(drug_ids):
        file_name = os.path.join(kegg_mol_files_dir, "%s.mol" % drug_id)
        if not os.path.exists(file_name):
            with open(file_name, "w") as mol_file_handler:
                kegg_mol_data = kegg_client.get(drug_id, 'mol')
                if isinstance(kegg_mol_data, int) and kegg_mol_data == 404:
                    not_found.append(drug_id)
                elif len(kegg_mol_data.strip()) == 0:
                    not_found.append(drug_id)
                else:
                    mol_file_handler.write(kegg_mol_data)
        yield i

    print("Not Found: %s" % (", ".join(not_found)))


def retrieve_zinc_properties(dest=os.path.join(data_dir, "zinc_16_prop.tsv")):
    """
    Retrieves ZINC properties file:
    "All Clean
    As Subset #6, but without 'yuck' compounds"

    """
    urlretrieve(ZINC_BASE_URL + "db/bysubset/16/16_prop.xls", dest)


def retrieve_zinc_structures(dest=os.path.join(data_dir, "zinc_16.sdf.gz")):
    """
    Retrieves ZINC structures file:
    "All Clean
    As Subset #6, but without 'yuck' compounds"

    """
    pbar = ProgressBar(maxval=len(ZINC_STRUCTURES), widgets=["Downloading Zinc Structures 16: ", Bar(), ETA()])
    with open(dest, 'wb') as output_file:
        for sdf_file in pbar(ZINC_STRUCTURES):
            response = requests.get(ZINC_SUBSET_16_BASE + "/" + sdf_file)
            output_file.write(response.content)



# Copyright 2017 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from marsi.chemistry import openbabel
from marsi.chemistry import rdkit


VALID_FP_FORMATS = openbabel.fps + rdkit.fps


class Molecule(object):
    """
    Object representing a molecule.

    """

    @classmethod
    def from_sdf(cls, path_or_str):
        """
        Builds a molecule from a file or string.

        Parameters
        ----------
        path_or_str : str
            The input file name or a string with the molecule description in SDF format.

        Returns
        -------
        marsi.chemistry.molecule.Molecule

        """
        from_file = os.path.isfile(path_or_str)
        ob_mol = openbabel.sdf_to_molecule(path_or_str, from_file)
        ob_mol.title = ""
        rd_mol = rdkit.sdf_to_molecule(path_or_str, from_file)
        return cls(ob_mol, rd_mol)

    @classmethod
    def from_inchi(cls, inchi):
        """
        Builds a molecule from an InChI key.

        Parameters
        ----------
        inchi : str
            A valid InChI key.

        Returns
        -------
        marsi.chemistry.molecule.Molecule

        """
        ob_mol = openbabel.inchi_to_molecule(inchi)
        rd_mol = rdkit.inchi_to_molecule(inchi)
        return cls(ob_mol, rd_mol)

    @classmethod
    def from_mol(cls, path_or_str):
        """
        Builds a molecule from a file or string.

        Parameters
        ----------
        path_or_str : str
            The input file name or a string with the molecule description in MOL format.

        Returns
        -------
        marsi.chemistry.molecule.Molecule

        """
        from_file = os.path.isfile(path_or_str)
        ob_mol = openbabel.mol_to_molecule(path_or_str, from_file)
        rd_mol = rdkit.mol_to_molecule(path_or_str, from_file)
        return cls(ob_mol, rd_mol)

    def __init__(self, ob_mol, rd_mol):
        self._ob_mol = ob_mol
        self._rd_mol = rd_mol

    @property
    def inchi(self):
        return openbabel.mol_to_inchi(self._ob_mol)

    @property
    def inchi_key(self):
        return openbabel.mol_to_inchi_key(self._ob_mol)

    @property
    def num_atoms(self):
        return self._ob_mol.OBMol.NumAtoms()

    @property
    def num_bonds(self):
        return self._ob_mol.OBMol.NumBonds()

    @property
    def num_rings(self):
        return len(self._ob_mol.OBMol.GetLSSR())

    def fingerprint(self, fpformat='maccs', bits=None):
        if fpformat not in VALID_FP_FORMATS:
            raise ValueError("Fingerprint '%s' is not valid. Use of of %s" % (fpformat, ", ".join(VALID_FP_FORMATS)))

        if fpformat in openbabel.fps:
            fp = openbabel.fingerprint(self._ob_mol, fpformat)
            bits = openbabel.fp_bits.get(fpformat, max(fp.bits))
            return openbabel.fingerprint_to_bits(fp, bits=bits)
        else:
            fp = rdkit.fingerprint(self._rd_mol, fpformat)
            if bits is None:
                bits = fp.GetNumBits()

            return rdkit.fingerprint_to_bits(fp, bits=bits)

    def _repr_html_(self):
        self._ob_mol.removeh()
        representation = self._ob_mol._repr_html_() or openbabel.mol_to_svg(self._ob_mol)
        self._ob_mol.addh()
        return representation
